#ifndef __GM_AUTH_H__
#define __GM_AUTH_H__

#include "../std.h"

#include <string>
#include <list>

#include <gssapi.h>

#include "../external/voms/api/voms_api.h"

#define AAA_POSITIVE_MATCH 1
#define AAA_NEGATIVE_MATCH -1
#define AAA_NO_MATCH 0
#define AAA_FAILURE 2

class AuthUser {
 private:
  typedef int (AuthUser:: * match_func_t)(const char* line);
  typedef struct {
    const char* cmd;
    match_func_t func;
  } source_t;
  string subject;
  string filename;
  bool proxy_file_was_created;
  static source_t sources[];
  int match_all(const char* line);
  int match_subject(const char* line);
  int match_file(const char* line);
  int match_ldap(const char* line);
  int match_voms(const char* line);
  vector<struct voms> voms_data;
 public:
  AuthUser(const AuthUser&);
  AuthUser(const char* subject = NULL,const char* filename = NULL);
  ~AuthUser(void);
  void operator=(const AuthUser&);
  void operator=(gss_cred_id_t cred);
  void set(const char* subject,gss_cred_id_t cred);
  int evaluate(const char* line);
  const char* DN(void) const { return subject.c_str(); };
  const char* proxy(void) const { return filename.c_str(); };
};

class AuthEvaluator {
 private:
  list<string> l;
  string name;
 public:
  AuthEvaluator(void);
  AuthEvaluator(const char* name);
  ~AuthEvaluator(void);
  void add(const char*);
  int evaluate(AuthUser &) const;
  bool operator==(const char* s) { return (strcmp(name.c_str(),s)==0); };
  bool operator==(const string& s) { return (name == s); };
};

#endif 
