#include "../std.h"

#include "../misc/url_options.h"
#include "../misc/stringtoint.h"
#include "../misc/globus_error_utils.h"
#include "../misc/globus_modules.h"
#include "../misc/log_time.h"
#include "datapoint.h"
#include "datahandle.h"

bool DataHandle::analyze(long int *bufsize,int *bufnum,bool *cache,bool *local) {
  if(!url) return false;
  long int bufsize_ = -1;
  long int bufnum_  = -1;
  const char* cur_url = url->current_location().c_str();
  const string* cur_url_s = &(url->current_location());
  string value;
  if(!strncasecmp("gsiftp://",cur_url,9)) {
    unsigned int threads;
    if(get_url_option(*cur_url_s,"threads",0,value) == 0) {
      if(stringtoint(value,threads)) {
        if(threads < 1) threads = 1;
        if(threads > MAX_PARALLEL_STREAMS) threads = MAX_PARALLEL_STREAMS;
      }
      else { threads = 1; };
      bufnum_=threads;
    };
  };
  bool cache_=true;
  if(get_url_option(*cur_url_s,"cache",0,value) == 0) {
    if(strcasecmp(value.c_str(),"no") == 0) cache_=false;
  };
  bool local_=false;
  if(!strncasecmp("file:/",cur_url,6)) {
    local_=true;
  };
  if(bufsize) (*bufsize)=bufsize_;
  if(bufnum) (*bufnum)=bufnum_;
  if(cache) (*cache)=cache_;
  if(local) (*local)=local_;
  return true;
}

bool DataHandle::deinit_handle(void) {
  if(ftp_active) {
    globus_ftp_client_handle_destroy(&ftp_handle);
    globus_ftp_client_operationattr_destroy(&ftp_opattr);
//    globus_ftp_client_plugin_destroy(&ftp_plugin);
  };
  if(http_active) {
    if(http_handle_was_initialized) {
      odlog(2)<<"DataHandle::deinit_handle: destroy http_handle"<<endl;
      globus_gass_transfer_request_destroy(http_handle);
    };
  };
  return true;
}

bool DataHandle::init_handle(void) {
  if(!url) return false;
  const char* cur_url = url->current_location().c_str();
  const string* cur_url_s = &(url->current_location());
  cacheable=true;
  string value;
  if(get_url_option(*cur_url_s,"cache",0,value) == 0) {
    if(strcasecmp("no",value.c_str()) == 0) cacheable=false;
  };
  c_url=cur_url;
  out_of_order(out_of_order());
  if(::canonic_url(c_url) != 0) return false;
  if((!strncasecmp("ftp://",cur_url,6)) || 
     (!strncasecmp("gsiftp://",cur_url,9))) {
    url_proto=url_is_ftp;
    is_secure=false;
    if(!strncasecmp("gsiftp://",cur_url,9)) is_secure=true;
    if(!ftp_active) {
      globus_result_t res;
      if((res=globus_ftp_client_handle_init(
                           &ftp_handle,GLOBUS_NULL)) != GLOBUS_SUCCESS) {
        odlog(0)<<"init_handle: globus_ftp_client_handle_init failed"<<endl;
        odlog(0)<<"Globus error: "<<res<<endl;
        ftp_active=false; return false;
      };
      if((res=globus_ftp_client_operationattr_init(
                           &ftp_opattr)) != GLOBUS_SUCCESS) {
        odlog(0)<<"init_handle: globus_ftp_client_operationattr_init failed"<<endl;
        odlog(0)<<"Globus error: "<<res<<endl;
        globus_ftp_client_handle_destroy(&ftp_handle);
        ftp_active=false; return false;
      };
//      if((res=globus_ftp_client_plugin_init(&ftp_plugin,"allo plugin",
//                 GLOBUS_FTP_CLIENT_CMD_MASK_FILE_ACTIONS,(void*)this)) 
//                                                       != GLOBUS_SUCCESS) {
//        odlog(0)<<"Globus error: "<<res<<endl;
//      } else {
//        globus_ftp_client_plugin_set_destroy_func(&ftp_plugin,&allo_plugin_destroy);
//        globus_ftp_client_plugin_set_copy_func(&ftp_plugin,&allo_plugin_copy);
//        globus_ftp_client_plugin_set_put_func(&ftp_plugin,&allo_plugin_put);
//        globus_ftp_client_plugin_set_command_func(&ftp_plugin,&allo_plugin_command);
//        globus_ftp_client_handle_add_plugin(&ftp_handle,&ftp_plugin);
//      };
    };
    ftp_active=true; ftp_threads=1;
    if(allow_out_of_order) {
      if(get_url_option(*cur_url_s,"threads",0,value) == 0) {
        if(stringtoint(value,ftp_threads)) {
          if(ftp_threads < 1) ftp_threads = 1;
          if(ftp_threads > MAX_PARALLEL_STREAMS) ftp_threads = MAX_PARALLEL_STREAMS;
        }
        else { ftp_threads = 1; };
      };
    };
    globus_ftp_control_parallelism_t paral;
    if(ftp_threads > 1) {
      paral.fixed.mode=GLOBUS_FTP_CONTROL_PARALLELISM_FIXED;
      paral.fixed.size=ftp_threads;
    } 
    else {
      paral.fixed.mode=GLOBUS_FTP_CONTROL_PARALLELISM_NONE;
      paral.fixed.size=1;
    };
    globus_ftp_client_operationattr_set_parallelism(&ftp_opattr,&paral);
    globus_ftp_client_operationattr_set_striped(&ftp_opattr,GLOBUS_FALSE);
    /*   globus_ftp_client_operationattr_set_layout         */
    /*   globus_ftp_client_operationattr_set_tcp_buffer     */
    globus_ftp_client_operationattr_set_type(&ftp_opattr,
                                    GLOBUS_FTP_CONTROL_TYPE_IMAGE);
    if(!is_secure) {
      globus_ftp_client_operationattr_set_mode(&ftp_opattr,
                                    GLOBUS_FTP_CONTROL_MODE_STREAM);
      globus_ftp_client_operationattr_set_data_protection(&ftp_opattr,
                                    GLOBUS_FTP_CONTROL_PROTECTION_CLEAR);
      globus_ftp_client_operationattr_set_control_protection(&ftp_opattr,
                                    GLOBUS_FTP_CONTROL_PROTECTION_CLEAR);
    }
    else {
      if(get_url_option(*cur_url_s,"secure",0,value) == 0) {
        if(value == "yes") {
          globus_ftp_client_operationattr_set_mode(&ftp_opattr,
                                    GLOBUS_FTP_CONTROL_MODE_EXTENDED_BLOCK);
          globus_ftp_client_operationattr_set_data_protection(&ftp_opattr,
                                      GLOBUS_FTP_CONTROL_PROTECTION_PRIVATE);
          odlog(2)<<"Using secure data transfer (requested in url)"<<endl;
        } else {
          if(force_passive) {
            globus_ftp_client_operationattr_set_mode(&ftp_opattr,
                                    GLOBUS_FTP_CONTROL_MODE_STREAM);
          } else {
            globus_ftp_client_operationattr_set_mode(&ftp_opattr,
                                    GLOBUS_FTP_CONTROL_MODE_EXTENDED_BLOCK);
          };
          globus_ftp_client_operationattr_set_data_protection(&ftp_opattr,
                                      GLOBUS_FTP_CONTROL_PROTECTION_CLEAR);
          odlog(2)<<"Using insecure data transfer (requested in url)"<<endl;
        };
      } else {
        if(force_secure) {
          globus_ftp_client_operationattr_set_mode(&ftp_opattr,
                                    GLOBUS_FTP_CONTROL_MODE_EXTENDED_BLOCK);
          globus_ftp_client_operationattr_set_data_protection(&ftp_opattr,
                                      GLOBUS_FTP_CONTROL_PROTECTION_PRIVATE);
        } else {
          if(force_passive) {
            globus_ftp_client_operationattr_set_mode(&ftp_opattr,
                                    GLOBUS_FTP_CONTROL_MODE_STREAM);
          } else {
            globus_ftp_client_operationattr_set_mode(&ftp_opattr,
                                    GLOBUS_FTP_CONTROL_MODE_EXTENDED_BLOCK);
          };
          globus_ftp_client_operationattr_set_data_protection(&ftp_opattr,
                                      GLOBUS_FTP_CONTROL_PROTECTION_CLEAR);
          odlog(2)<<"Using insecure data transfer"<<endl;
        };
      };
      globus_ftp_client_operationattr_set_control_protection(&ftp_opattr,
                                    GLOBUS_FTP_CONTROL_PROTECTION_PRIVATE);
    };
    /*   globus_ftp_client_operationattr_set_dcau                         */
    /*   globus_ftp_client_operationattr_set_resume_third_party_transfer  */
    /*   globus_ftp_client_operationattr_set_authorization                */
    globus_ftp_client_operationattr_set_append(&ftp_opattr,GLOBUS_FALSE);
    return true;
  }
  else if((!strncasecmp("http://",cur_url,7)) || 
          (!strncasecmp("https://",cur_url,8))) {
    url_proto=url_is_http;
    if(!http_active) {
      http_handle_was_initialized=false;
      http_active=true;
    };
    return true;
  }
  else if((!strncasecmp("bbftp://",cur_url,8)) || 
          (!strncasecmp("gsibbftp://",cur_url,11))) {
    url_proto=url_is_bbftp;



    return false;
  }
  else if(!strncasecmp("file:/",cur_url,6)) {
    url_proto=url_is_file;


    return true;
  };
  return false;
}

void DataHandle::allo_plugin_command(
    globus_ftp_client_plugin_t *                plugin,
    void *                                      plugin_specific,
    globus_ftp_client_handle_t *                handle,
    const char *                                url,
    const char *                                command) {
  DataHandle* it = (DataHandle*)plugin_specific;
  if((strncasecmp(command,"STOR",4) == 0) ||
     (strncasecmp(command,"ESTO",4) == 0)) {
    if(it->url->meta_size_available() && (it->url->meta_size() > 0)) {
      /* Send ALLO to try to preallocate or at least check space on server */
      /*
        I need a way to get globus_ftp_control_hnadle_t from globus_ftp_client_handle_t
      */
    };
  };
}

void DataHandle::allo_plugin_put(
    globus_ftp_client_plugin_t *                plugin,
    void *                                      plugin_specific,
    globus_ftp_client_handle_t *                handle,
    const char *                                url,
    const globus_ftp_client_operationattr_t *   attr,
    globus_bool_t                               restart) {
}

globus_ftp_client_plugin_t* DataHandle::allo_plugin_copy(
    globus_ftp_client_plugin_t *                plugin_template,
    void *                                      plugin_specific) {
  globus_result_t res;
  DataHandle* it = (DataHandle*)plugin_specific;
  globus_ftp_client_plugin_t* new_plugin = (globus_ftp_client_plugin_t *)
        globus_malloc(sizeof(globus_ftp_client_plugin_t));

  if(new_plugin == GLOBUS_NULL) return GLOBUS_NULL;

  if((res=globus_ftp_client_plugin_init(&(it->ftp_plugin),"allo plugin",
             GLOBUS_FTP_CLIENT_CMD_MASK_FILE_ACTIONS,plugin_specific))
                                                   != GLOBUS_SUCCESS) {
    odlog(0)<<"Globus error: "<<res<<endl;
    globus_free(new_plugin);
    return GLOBUS_NULL;
  } else {
    globus_ftp_client_plugin_set_destroy_func(&(it->ftp_plugin),&allo_plugin_destroy);
    globus_ftp_client_plugin_set_copy_func(&(it->ftp_plugin),&allo_plugin_copy);
    globus_ftp_client_plugin_set_put_func(&(it->ftp_plugin),&allo_plugin_put);
    globus_ftp_client_plugin_set_command_func(&(it->ftp_plugin),&allo_plugin_command);
  };
  return new_plugin;
}

void DataHandle::allo_plugin_destroy(
  globus_ftp_client_plugin_t *                plugin,
  void *                                      plugin_specific) {
  globus_ftp_client_plugin_destroy(plugin);
  globus_free(plugin);
}

DataHandle::DataHandle(DataPoint* url_) {
  url=url_;
  /* connection is not initialised here - only if request is made */
  reading=false;
  writing=false;
  ftp_active=false; http_active=false; bbftp_active=false;
  if(!ftp_mod.active()) {
    odlog(0)<<"GLOBUS_FTP_CLIENT_MODULE activation failed"<<endl;
    url=NULL;
  };
  if(!http_mod.active()) {
    odlog(0)<<"GLOBUS_GASS_TRANSFER_MODULE activation failed"<<endl;
    url=NULL;
  };
  is_secure=false; force_secure=true; force_passive=false;
}

DataHandle::~DataHandle(void) {
  /*  Stop any transfer */
  stop_reading();
  stop_writing();
  /* !!!!!!!!!!! destroy activated handles, etc here !!!!!!!! */
  deinit_handle();
}

bool DataHandle::start_reading(DataBufferPar &buffer) {
  if(reading || writing || (url==NULL)) return false;
  if(!init_handle()) return false;
  reading=true;
  switch(url_proto) {
    case url_is_file: {
      if(!start_reading_file(buffer)) { reading=false; return false; };
    }; break;
    case url_is_ftp: {
      if(!start_reading_ftp(buffer)) { reading=false; return false; };
    }; break;
    case url_is_http: {
      if(!start_reading_http(buffer)) { reading=false; return false; };
    }; break;
    case url_is_bbftp: {
      return false;
    }; break;
    default: reading=false; return false;
  };  
  return true;
}

bool DataHandle::check(void) {
  if(reading || writing || (url==NULL)) return false;
  if(!init_handle()) return false;
  switch(url_proto) {
    case url_is_file: {
      if(!check_file()) { return false; };
    }; break;
    case url_is_ftp: {
      if(!check_ftp()) { return false; };
    }; break;
    case url_is_http: {
      if(!check_http()) { return false; };
    }; break;
    case url_is_bbftp: {
      return false;
    }; break;
    default: return false;
  };
  return true;
}

bool DataHandle::remove(void) {
  if(reading || writing || (url==NULL)) return false;
  if(!init_handle()) return false;
  switch(url_proto) {
    case url_is_file: {
      if(!remove_file()) { return false; };
    }; break;
    case url_is_ftp: {
      if(!remove_ftp()) { return false; };
    }; break;
    case url_is_http: {
      return false;
    }; break;
    case url_is_bbftp: {
      return false;
    }; break;
    default: return false;
  };
  return true;
}

bool DataHandle::start_writing(DataBufferPar &buffer,DataCallback *space_cb) {
  if(reading || writing || (url==NULL)) return false;
  if(!init_handle()) return false;
  writing=true;
  switch(url_proto) {
    case url_is_file: {
      if(!start_writing_file(buffer,space_cb)) { writing=false; return false; };
    }; break;
    case url_is_ftp: {
      if(!start_writing_ftp(buffer)) { writing=false; return false; };
    }; break;
    case url_is_http: {
      if(!start_writing_http(buffer)) { writing=false; return false; };
    }; break;
    case url_is_bbftp: {
      return false;
    }; break;
    default: writing=false; return false;
  };
  return true;
}

bool DataHandle::stop_reading(void) {
  if(!reading) return false;
  reading=false;   
  switch(url_proto) {
    case url_is_file: {
      return stop_reading_file();
    }; break;
    case url_is_ftp: {
      return stop_reading_ftp();
    }; break;
    case url_is_http: {
      return stop_reading_http();
    }; break;
    case url_is_bbftp: {
    }; break;
  };
  return true;
}

bool DataHandle::stop_writing(void) {
  if(!writing) return false;
  writing=false;
  switch(url_proto) {
    case url_is_file: {
      return stop_writing_file();
    }; break;
    case url_is_ftp: {
      return stop_writing_ftp();
    }; break;
    case url_is_http: {
      return stop_writing_http();
    }; break;
    case url_is_bbftp: {
    }; break;
  };
  return true;
}

bool DataHandle::out_of_order(void) {
  const char* cur_url = url->current_location().c_str();
  if((strncasecmp(cur_url,"gsiftp://",9)==0) || (strncasecmp(cur_url,"file:/",6)==0)) return true;
  return false;
}

void DataHandle::out_of_order(bool val) {
  allow_out_of_order=val;
}

bool DataHandle::secure(void) {
  return is_secure;
}

void DataHandle::secure(bool val) {
  force_secure=val;
}

void DataHandle::passive(bool val) {
  force_passive=val;
}

