#include "../std.h"

#ifdef HAVE_RLS_CLIENT
extern "C" {
#include <globus_rls_client.h>
}
#endif

#include "../config/config_map.h"
#include "../transfer/replica_utils.h"
#include "../transfer/replica.h"
#include "../misc/url_options.h"
#include "../misc/checksum.h"
#include "../misc/stringtoint.h"
#include "../misc/inttostring.h"
#include "../misc/log_time.h"

#include "datapoint.h"

#ifdef HAVE_RLS_CLIENT
static int globus_rls_client_lrc_attr_put(globus_rls_handle_t* h,char* key,globus_rls_attribute_t *attr,int overwrite) {
  int err;
  err = globus_rls_client_lrc_attr_add(h,key,attr);
  if(err != GLOBUS_RLS_SUCCESS) {
    if((overwrite) && (err == GLOBUS_RLS_DBERROR)) { /* guess this can mean duplicate entry */
      int err_;
      err_ = globus_rls_client_lrc_attr_remove(h,key,attr);
      if(err_ != GLOBUS_RLS_SUCCESS) return err;
      return globus_rls_client_lrc_attr_put(h,key,attr,0);
    };
    if(err != GLOBUS_RLS_ATTR_NEXIST) return err;
    err = globus_rls_client_lrc_attr_create(h,attr->name,
                                               attr->objtype,attr->type);
    if(err != GLOBUS_RLS_SUCCESS) return err;
    err = globus_rls_client_lrc_attr_add(h,key,attr);
  };
  return err;
}
#endif

/* perform resolve operation, which can take long time */
bool DataPoint::meta_resolve_rls(bool source) {
#ifdef HAVE_RLS_CLIENT
  is_resolved=false;
  if(!source) { /* destination can't derive locations - 
                       they must be given */
    odlog(1)<<"Locations are missing is destination RLS url"<<endl;
    if(locations.size() == 0) return false;
  };
  is_metaexisting=false;

  globus_rls_handle_t* h;
  globus_list_t* pfns_list;
  char errmsg[MAXERRMSG];
  int err;

  err = globus_rls_client_connect((char*)(rls_url.c_str()),&h,errmsg,MAXERRMSG);
  if(err != GLOBUS_RLS_SUCCESS) {
    odlog(1)<<"Failed to connecct RLS server: "<<
            globus_rls_client_errmsg(h,err,NULL,0)<<endl;
    odlog(1)<<"server error: "<<errmsg<<endl;
  };
#if (RLS_VERSION_MAJOR < 0) || ((RLS_VERSION_MAJOR == 1) && (RLS_VERSION_MINOR <1))
  err=globus_rls_client_lrc_get_pfn(h,(char*)(rls_lfn.c_str()),&pfns_list);
#else
  err=globus_rls_client_lrc_get_pfn(h,(char*)(rls_lfn.c_str()),0,0,&pfns_list);
#endif

  if(err != GLOBUS_RLS_SUCCESS) {
    if(err == GLOBUS_RLS_LFN_NEXIST) {
      globus_rls_client_close(h);
      if(source) {
        odlog(1)<<"Requested LFN (for source) is missing"<<endl;
        return false;
      };
      /* ok for destination */
      is_resolved=true;
      return true;
    };
    odlog(1)<<"Failed to obtain PFNs: "<<
            globus_rls_client_errmsg(h,err,NULL,0)<<endl;
    globus_rls_client_close(h);
    return false;
  };
  is_metaexisting=true;
  if(source) {  /* TODO: remove existing pfns from locations for !source */
    if(locations.size() == 0) {
      globus_list_t* list_p;
      for(list_p=pfns_list;list_p;list_p=globus_list_rest(list_p)) {
        globus_rls_string2_t* str2=
                         (globus_rls_string2_t*)globus_list_first(list_p);
        char* pfn = str2->s1;
        locations.push_back(DataPoint::Location(pfn,pfn));
      }; 
    }
    else {
      list<Location>::iterator loc = locations.begin();
      for(;loc!=locations.end();) {
        bool found = false;
        globus_list_t* list_p;
        for(list_p=pfns_list;list_p;list_p=globus_list_rest(list_p)) {
          globus_rls_string2_t* str2=
                           (globus_rls_string2_t*)globus_list_first(list_p);
          char* pfn = str2->s1;
          if(strncmp(pfn,loc->meta.c_str(),loc->meta.length()) == 0) {
            found=true;
            break;
          };
          if(!found) { loc=locations.erase(loc); }
          else { loc->url=loc->meta; ++loc; };
        };
      };
    };
  };
  globus_rls_client_free_list(pfns_list);
  /* obtain metadata */
  globus_list_t* attr_list;
  err = globus_rls_client_lrc_attr_value_get(h,(char*)(rls_lfn.c_str()),NULL,
                                  globus_rls_obj_lrc_lfn,&attr_list);
  if(err != GLOBUS_RLS_SUCCESS) {
    if(err == GLOBUS_RLS_ATTR_NEXIST) { 
      globus_rls_client_close(h);
      is_resolved=true;
      return true;
    };
    odlog(1)<<"Failed to obtain attributes: "<<
            globus_rls_client_errmsg(h,err,NULL,0)<<endl;
    globus_rls_client_close(h);
    return false;
  };
  globus_rls_client_close(h);
  for(globus_list_t* list_pa=attr_list;list_pa;
                                       list_pa=globus_list_rest(list_pa)) {
     unsigned long long int i;
    globus_rls_attribute_t* attr = 
                          (globus_rls_attribute_t*)globus_list_first(list_pa);
    if(attr->type != globus_rls_attr_type_str) continue;
    if(strcmp(attr->name,"filechecksum") == 0) {
      if(!meta_checksum_valid) if(stringtoint(attr->val.s,i)) meta_checksum(i);
    }
    else if(strcmp(attr->name,"size") == 0) {
      if(!meta_size_valid) if(stringtoint(attr->val.s,i)) meta_size(i);
    }
    else if(strcmp(attr->name,"modifytime") == 0) {
      if(!meta_created_valid) if(stringtoint(attr->val.s,i)) meta_created(i);
    };
  };
  globus_rls_client_free_list(attr_list);
  odlog(2)<<"meta_get_data: checksum: "<<meta_checksum()<<endl;
  odlog(2)<<"meta_get_data: size: "<<meta_size()<<endl;
  odlog(2)<<"meta_get_data: created: "<<meta_created()<<endl;
  location=locations.begin();
  is_resolved=true;
  return true;
#else
  return false;
#endif
}

bool DataPoint::meta_preregister_rls(bool replication) {
#ifdef HAVE_RLS_CLIENT
  if(replication) { /* replicating inside same lfn */
    if(!is_metaexisting) { /* for replication it must be there */
      odlog(0)<<"LFN is missing in RLS (needed for replication)"<<endl;
      return false;
    };
    return true;
  };
  if(is_metaexisting) { /* algorithm require this to be new file */
    odlog(0)<<"LFN already exists in replica"<<endl;
    return false;
  };
  /* RLS does not support LFN only in database - hence doing nothing here */
  return true;
#else
  return false;
#endif
}

bool DataPoint::meta_postregister_rls(bool replication,bool failure) {
#ifdef HAVE_RLS_CLIENT
  globus_rls_handle_t* h;
  char errmsg[MAXERRMSG];
  int err;

  err = globus_rls_client_connect((char*)(rls_url.c_str()),&h,errmsg,MAXERRMSG);  if(err != GLOBUS_RLS_SUCCESS) {
    odlog(1)<<"Failed to connecct RLS server: "<<
            globus_rls_client_errmsg(h,err,NULL,0)<<endl;
    odlog(1)<<"server error: "<<errmsg<<endl;
  };
  if(!replication) {
    if((err = globus_rls_client_lrc_create(h,(char*)(rls_lfn.c_str()),
                  (char*)(location->url.c_str()))) != GLOBUS_RLS_SUCCESS) {
      if(err == GLOBUS_RLS_LFN_EXIST) {
         err = globus_rls_client_lrc_add(h,(char*)(rls_lfn.c_str()),
                                           (char*)(location->url.c_str()));
      };
    };
  }
  else {
    err = globus_rls_client_lrc_add(h,(char*)(rls_lfn.c_str()),
                                           (char*)(location->url.c_str()));
  };
  if((err != GLOBUS_RLS_SUCCESS) && (err != GLOBUS_RLS_MAPPING_EXIST)) {
    odlog(1)<<"Failed to create/add LFN-PFN mapping: "<<
            globus_rls_client_errmsg(h,err,NULL,0)<<endl;
    globus_rls_client_close(h);
    return false;
  };
  globus_rls_attribute_t attr;
  string attr_val;
  attr.objtype=globus_rls_obj_lrc_lfn;
  attr.type=globus_rls_attr_type_str;
  attr.name="filetype";
  attr.val.s="file";
  err=globus_rls_client_lrc_attr_put(h,(char*)(rls_lfn.c_str()),&attr,0);
  if((err != GLOBUS_RLS_SUCCESS) && (err != GLOBUS_RLS_ATTR_EXIST)) {
    odlog(1)<<"Warning: failed to add attribute to RLS: "<<
                     globus_rls_client_errmsg(h,err,NULL,0)<<endl;
  };
  if(meta_size_valid) {
    attr.name="size";
    attr_val=inttostring(meta_size_);
    attr.val.s=(char*)(attr_val.c_str());
    err=globus_rls_client_lrc_attr_put(h,(char*)(rls_lfn.c_str()),&attr,0);
    if((err != GLOBUS_RLS_SUCCESS) && (err != GLOBUS_RLS_ATTR_EXIST)) {
      odlog(1)<<"Warning: failed to add attribute to RLS: "<<
                       globus_rls_client_errmsg(h,err,NULL,0)<<endl;
    };
  };
  if(meta_checksum_valid) {
    attr.name="filechecksum";
    attr_val=inttostring(meta_checksum_);
    attr.val.s=(char*)(attr_val.c_str());
    err=globus_rls_client_lrc_attr_put(h,(char*)(rls_lfn.c_str()),&attr,0);
    if((err != GLOBUS_RLS_SUCCESS) && (err != GLOBUS_RLS_ATTR_EXIST)) {
      odlog(1)<<"Warning: failed to add attribute to RLS: "<<
                       globus_rls_client_errmsg(h,err,NULL,0)<<endl;
    };
  };
  if(meta_created_valid) {
    attr.name="modifytime";
    attr_val=inttostring(meta_created_);
    attr.val.s=(char*)(attr_val.c_str());
    err=globus_rls_client_lrc_attr_put(h,(char*)(rls_lfn.c_str()),&attr,0);
    if((err != GLOBUS_RLS_SUCCESS) && (err != GLOBUS_RLS_ATTR_EXIST)) {
      odlog(1)<<"Warning: failed to add attribute to RLS: "<<
                       globus_rls_client_errmsg(h,err,NULL,0)<<endl;
    };
  };
  globus_rls_client_close(h);
  return true;
#else
  return false;
#endif
}

bool DataPoint::meta_preunregister_rls(bool replication) {
#ifdef HAVE_RLS_CLIENT
  if(replication) return true;
  return true;
#else
  return false;
#endif
}

bool DataPoint::meta_unregister_rls(bool all) {
#ifdef HAVE_RLS_CLIENT
  if(!all) {
    if(location == locations.end()) { 
      odlog(0)<<"Location is missing"<<endl;
      return false;
    };
  };
  globus_rls_handle_t* h;
  char errmsg[MAXERRMSG];
  int err;
  globus_list_t* pfns_list;

  err = globus_rls_client_connect((char*)(rls_url.c_str()),&h,errmsg,MAXERRMSG);
  if(err != GLOBUS_RLS_SUCCESS) {
    odlog(1)<<"Failed to connect RLS server: "<<
            globus_rls_client_errmsg(h,err,NULL,0)<<endl;
    odlog(1)<<"server error: "<<errmsg<<endl;
  };
  if(all) {
    /* get all pfns and delete them */
#if (RLS_VERSION_MAJOR < 0) || ((RLS_VERSION_MAJOR == 1) && (RLS_VERSION_MINOR <1))
    err = globus_rls_client_lrc_get_pfn(h,(char*)(rls_lfn.c_str()),&pfns_list);
#else
    err = globus_rls_client_lrc_get_pfn(h,(char*)(rls_lfn.c_str()),0,0,&pfns_list);
#endif
    if(err != GLOBUS_RLS_SUCCESS) {
      if(err == GLOBUS_RLS_LFN_NEXIST) {
        odlog(1)<<"Warning: Requested LFN does not already exist"<<endl;
        globus_rls_client_close(h);
        return true;
      };
      odlog(1)<<"Failed to obtain PFNs: "<<
            globus_rls_client_errmsg(h,err,NULL,0)<<endl;
      globus_rls_client_close(h);
      return false;
    };
    globus_list_t* list_p;
    for(list_p=pfns_list;list_p;list_p=globus_list_rest(list_p)) {
      globus_rls_string2_t* str2=
                       (globus_rls_string2_t*)globus_list_first(list_p);
      char* pfn = str2->s1;
      err = globus_rls_client_lrc_delete(h,(char*)(rls_lfn.c_str()),pfn);
      if((err != GLOBUS_RLS_SUCCESS) && (err != GLOBUS_RLS_MAPPING_NEXIST)) {
        odlog(1)<<"Failed to delete LFN-PFN mapping: "<<
                globus_rls_client_errmsg(h,err,NULL,0)<<endl;
        globus_rls_client_close(h);
        globus_rls_client_free_list(pfns_list);
        return false;
      };
    };
    globus_rls_client_free_list(pfns_list);
  }
  else {
    err = globus_rls_client_lrc_delete(h,(char*)(rls_lfn.c_str()),(char*)(location->url.c_str()));
    if((err != GLOBUS_RLS_SUCCESS) && (err != GLOBUS_RLS_MAPPING_NEXIST)) {
      odlog(1)<<"Failed to delete LFN-PFN mapping: "<<
              globus_rls_client_errmsg(h,err,NULL,0)<<endl;
      globus_rls_client_close(h);
      return false;
    };
  };
  globus_rls_client_close(h);
  return true;
#else
  return false;
#endif
}

/* rls://[url[|url[...]]@]server/lfn */
bool DataPoint::process_meta_url_rls(const string &url,string &cat_url,string &filename,list<DataPoint::Location> &locations) {
  if(!strncasecmp(url.c_str(),"rls://",6)) {
    cat_url.resize(0); locations.clear(); filename.resize(0);
    string url_(url.c_str());
    /* find out if it contains locations */
    int loc_start=6;
    int loc_end=url_.find('@',loc_start);
    string urls("");
    if(loc_end!=string::npos) {
      urls=url_.substr(loc_start,loc_end-loc_start);
      url_.erase(loc_start,loc_end-loc_start+1);
    };
    /* get lfn */
    int server_start=6;
    int server_end=url_.find('/',loc_start);
    if(server_end==string::npos) {
      odlog(1)<<"LFN is missing in url "<<url_<<endl;
      return false;
    };
    filename=url_.substr(server_end+1);
    cat_url=url_.substr(0,server_end);
    odlog(2)<<"LFN: "<<filename<<endl;
    odlog(2)<<"RLS server: "<<cat_url<<endl;
    odlog(2)<<"Location urls: "<<urls<<endl;
    int n=0;
    for(int nn=0;n<urls.length();) {
      nn=urls.find('|',n); if(nn == string::npos) nn=urls.length();
      if(n == nn) { n++; continue; };
      string loc(urls.c_str()+n,nn-n);
      locations.push_back(DataPoint::Location(loc.c_str(),loc.c_str()));
      n=nn+1;
    };
    return true;
  };
  return false;
};

