/*********************************************************************
 *
 * Authors: Vincenzo Ciaschini - Vincenzo.Ciaschini@cnaf.infn.it 
 *
 * Copyright (c) 2002, 2003 INFN-CNAF on behalf of the EU DataGrid.
 * For license conditions see LICENSE file or
 * http://www.edg.org/license.html
 *
 * Parts of this code may be based upon or even include verbatim pieces,
 * originally written by other people, in which case the original header
 * follows.
 *
 *********************************************************************/
#include <stdio.h>
#include <voms_apic.h>

#include <openssl/pem.h>
#include <openssl/bio.h>
#include <openssl/x509.h>

static char *retmsg[] = { "VERR_NONE", "VERR_NOSOCKET", "VERR_NOIDENT", "VERR_COMM", 
			  "VERR_PARAM", "VERR_NOEXT", "VERR_NOINIT",
			  "VERR_TIME", "VERR_IDCHECK", "VERR_EXTRAINFO",
			  "VERR_FORMAT", "VERR_NODATA", "VERR_PARSE",
			  "VERR_DIR", "VERR_SIGN", "VERR_SERVER", 
			  "VERR_MEM", "VERR_VERIFY", "VERR_IDENT",
			  "VERR_TYPE" };

void print(struct vomsdata *d)
{
  struct voms **vo = d->data;
  struct voms *v;
  int k = 0;
  int j =0;
  
  while(vo[k]) {
    v = vo[k++];
    printf("%d *******************************************\n",k);
    printf("SIGLEN: %d\nUSER: %s\n", v->siglen, v->user);
    printf("UCA: %s\nSERVER: %s\n", v->userca, v->server);
    printf("SCA: %s\nVO: %s\n", v->serverca, v->voname);
    printf("URI: %s\nDATE1: %s\n", v->uri, v->date1);
    printf("DATE2: %s\n", v->date2);

    switch (v->type) {
    case TYPE_NODATA:
      printf("NO DATA\n");
      break;
    case TYPE_CUSTOM:
      printf("%*s\n", v->datalen - 10, v->custom);
      break;
    case TYPE_STD:
      j = 0;
      while (v->std[j]) {
	printf("GROUP: %s\nROLE: %s\nCAP: %s\n",v->std[j]->group,
	       v->std[j]->role,v->std[j]->cap);
	j++;
      }
    }
  }
  
  printf("WORKVO: %s\nEXTRA: %s\n", d->workvo,d->extra_data);
}

int main(int argc, char *argv[])
{
  struct vomsdata *vd = VOMS_Init("/home/marotta/cert", NULL);
  int err;
  int res;
  BIO *in = NULL;
  X509 *x = NULL;

  if (!vd) {
    printf("ERR:!\n");
    exit(0);
  }
  /*
   * Initializes directory info.
   */

  printf("TEST 1\n");
  /*
   * Contact two different vomses and collect information from both.
   */
  res = VOMS_Contact("aaa-test.cnaf.infn.it",15000,"/C=IT/O=INFN/OU=cas server/L=Bologna/CN=cas/aaa-test.cnaf.infn.it/Email=Vincenzo.Ciaschini@cnaf.infn.it", "S1",vd, &err);
  res = VOMS_Contact("datatag6.cnaf.infn.it",15000,"/C=IT/O=INFN/OU=cas server/L=Bologna/CN=cas/aaa-test.cnaf.infn.it/Email=Vincenzo.Ciaschini@cnaf.infn.it", "A",vd, &err);

  if (res)
    print(vd);
  else
    printf("ERROR!\n");
  if (!res) {
    printf("err: %s\n", retmsg[err]);
  }

  printf("TEST 2\n");
  /*
   * Get Data from real certificate.
   */

  VOMS_Destroy(vd);
  vd = VOMS_Init("/home/marotta/cert", NULL);


  in = BIO_new(BIO_s_file());
  if (in) {
    if (BIO_read_filename(in, "/tmp/x509up_u500") > 0) {
      x = PEM_read_bio_X509(in, NULL, 0, NULL);
      res = VOMS_Retrieve(x, NULL, RECURSE_NONE, vd, &err);

      if (res)
	print(vd);
      else
	printf("ERROR!\n");
    }
  }

  if (!res) {
    printf("err: %s\n", retmsg[err]);
  }
  
  exit(0);
}
