#include "../std.h"
#include "fileroot.h"
#include "config.h"
#include "names.h"
#include "misc.h"
#include "../misc/canonical_dir.h"
#include "../misc/log_time.h"

string FileNode::last_name(void) {
  int pl=point.rfind('/');
  if(pl == -1) return point;
  return point.substr(pl+1);
}

bool FileNode::belongs(const char* name) {
  int pl=point.length();
  int l=strlen(name);
  if (pl > l) return false;
  if(strncmp(point.c_str(),name,pl)) return false;
  if(pl == l) return true;
  if(name[pl] == '/') return true;
  return false;
}

/*
bool FileNode::is_in_dir(const char* name) {
  int l=strlen(name);
  if(point.length() <= l) return false;
  if(point[l] != '/') return false;
  if(strncmp(point.c_str(),name,l)) return false;
  return true;
}
*/
/* should only last directory be shown ? */
bool FileNode::is_in_dir(const char* name) {
  int l=strlen(name);
  int pl=point.rfind('/'); /* returns -1 if not found */
  if(pl == -1) {
    if(l == 0) { 
      return true;
    };
  };
  if(l != pl) return false;
  if(strncmp(point.c_str(),name,l)) return false;
  return true;
}

int FileRoot::size(const char* name,unsigned long long int *size) {
  string new_name;
  if(name[0] != '/') { new_name=cur_dir+'/'+name; }
  else { new_name=name; };
  if(canonical_dir(new_name,false) != 0) return 1;
  for(list<FileNode>::iterator i=nodes.begin();i!=nodes.end();++i) {
    if(i->belongs(new_name.c_str())) {
      DirEntry info;
      if(i->checkfile(new_name,info,DirEntry::basic_object_info) != 0) return 1;
      (*size)=info.size; 
      return 0;
    };
  };
  return 1;
}

int FileRoot::time(const char* name,time_t *time) {
  string new_name;
  if(name[0] != '/') { new_name=cur_dir+'/'+name; }
  else { new_name=name; };
  if(canonical_dir(new_name,false) != 0) return 1;
  for(list<FileNode>::iterator i=nodes.begin();i!=nodes.end();++i) {
    if(i->belongs(new_name.c_str())) {
      DirEntry info;
      if(i->checkfile(new_name,info,DirEntry::basic_object_info) != 0) {
        return 1;
      };
      (*time)=info.modified; 
      return 0;
    };
  };
  return 1;
}

int FileRoot::checkfile(const char* name,DirEntry &info,DirEntry::object_info_level mode) {
  string new_name;
  if(name[0] != '/') { new_name=cur_dir+'/'+name; }
  else { new_name=name; };
  if(canonical_dir(new_name,false) != 0) return 1;
  for(list<FileNode>::iterator i=nodes.begin();i!=nodes.end();++i) {
    if(i->belongs(new_name.c_str())) {
      DirEntry info;
      if(i->checkfile(new_name,info,mode) != 0) {
        return 1;
      };
      info.name="/"+new_name;
      return 0;
    };
  };
  return 1;
}

int FileRoot::mkd(string& name) {
  string new_dir;
  if(name[0] != '/') { new_dir=cur_dir+'/'+name; }
  else { new_dir=name; };
  if(canonical_dir(new_dir,false) == 0) {
    for(list<FileNode>::iterator i=nodes.begin();i!=nodes.end();++i) {
      if((*i) == new_dir) { /* already exists, at least virtually */
        name=new_dir;
        return 0;
      };
      if((*i).belongs(new_dir.c_str())) {
        if((*i).makedir(new_dir) == 0) {
          name=new_dir;
          return 0;
        };
        name=cur_dir;
        return 1;
      };
    };
  };
  name=cur_dir;
  return 1;
}

int FileRoot::rmd(string& name) {
  string new_dir;
  if(name[0] != '/') { new_dir=cur_dir+'/'+name; }
  else { new_dir=name; };
  if(canonical_dir(new_dir,false) == 0) {
    for(list<FileNode>::iterator i=nodes.begin();i!=nodes.end();++i) {
      if((*i) == new_dir) { /* virtual - not removable */
        return 1;
      };
      if(i->belongs(new_dir.c_str())) {
        return i->removedir(new_dir);
      };
    };
  };
  return 1;
}

int FileRoot::rm(string& name) {
  string new_dir;
  if(name[0] != '/') { new_dir=cur_dir+'/'+name; }
  else { new_dir=name; };
  if(canonical_dir(new_dir,false) == 0) {
    for(list<FileNode>::iterator i=nodes.begin();i!=nodes.end();++i) {
      if((*i) == new_dir) { /* virtual dir - not removable */
        return 1;
      };
      if(i->belongs(new_dir.c_str())) {
        return i->removefile(new_dir);
      };
    };
  };
  return 1;
}

int FileRoot::cwd(string& name) {
  string new_dir;
  if(name[0] != '/') { new_dir=cur_dir+'/'+name; }
  else { new_dir=name; };
  if(canonical_dir(new_dir,false) == 0) {
    if(new_dir.length() == 0) { /* always can go to root ? */
      cur_dir=new_dir;
      name=cur_dir;
      return 0;
    };
    /* check if can cd */
    for(list<FileNode>::iterator i=nodes.begin();i!=nodes.end();++i) {
      if((*i) == new_dir) {
        cur_dir=new_dir;
        name=cur_dir;
        return 0;
      };
      if((*i).belongs(new_dir.c_str())) {
        if((*i).checkdir(new_dir) == 0) {
          cur_dir=new_dir;
          name=cur_dir;
          return 0;
        };
        name=cur_dir;
        return 1;
      };
    };
  };
  name=cur_dir;
  return 1;
}

int FileRoot::open(const char* name,open_modes mode,unsigned long long int size) {
  string new_name;
  if(name[0] != '/') { new_name=cur_dir+'/'+name; }
  else { new_name=name; };
  if(canonical_dir(new_name,false) != 0) { return 1; };
  for(list<FileNode>::iterator i=nodes.begin();i!=nodes.end();++i) {
    if(i->belongs(new_name.c_str())) {
      if(i->open(new_name.c_str(),mode,size) == 0) {
        opened_node=i;
        return 0;
      };
      return 1;
    };
  };
  return 1;
}

int FileRoot::close(bool eof) {
  if(opened_node != nodes.end()) {
    int i=(*opened_node).close(eof);
    opened_node=nodes.end();
    return i;
  };
  return 1;
}

int FileRoot::read(unsigned char* buf,unsigned long long int offset,unsigned long long *size) {
  if(opened_node != nodes.end()) {
    return (*opened_node).read(buf,offset,size);
  };
  return 1;
}

int FileRoot::write(unsigned char *buf,unsigned long long int offset,unsigned long long size) {
  if(opened_node != nodes.end()) {
    return (*opened_node).write(buf,offset,size);
  };
  return 1;
}

/* 0 - ok , 1 - failure, -1 - this is a file */
int FileRoot::readdir(const char* name,list<DirEntry> &dir_list,DirEntry::object_info_level mode) {
  string fullname;
  int offset=0;
  bool belongs_found = false;
  if(name[0] != '/') { fullname=cur_dir+'/'+name; }
  else { fullname=name; };
  if(canonical_dir(fullname,false) != 0) return 1;
  int res = 1;
  for(list<FileNode>::iterator i=nodes.begin();i!=nodes.end();++i) {
    if(i->belongs(fullname.c_str())) {
      belongs_found=true;
      res=i->readdir(fullname.c_str(),dir_list,mode);
      break;
    };
  };
  if(res == -1) { /* means this is a file */
    list<DirEntry>::iterator di = dir_list.end(); --di;
    di->name="/"+fullname;
    return -1;
  };
  for(list<FileNode>::iterator i=nodes.begin();i!=nodes.end();++i) {
    if(i->is_in_dir(fullname.c_str())) {
      DirEntry de;
      de.name=i->last_name();
      de.is_file=false;

//      if(i->checkfile(i->point,de,mode) == 0) {
//        if(de.is_file) {
//          de.reset(); de.name=i->last_name(); de.is_file=false;
          /* TODO: fill other attributes */
//
//        };
//      };
      dir_list.push_front(de);
      res=0;
    };
  };
  return res;
}

int FileRoot::config(globus_ftp_control_auth_info_t *auth,
                     gss_cred_id_t delegated_cred,
                     globus_ftp_control_handle_t *handle) {
  /* open and read configuration file */
  ifstream cfile;
  config_open(cfile);
  if(!cfile.is_open()) {
    olog << "Error: configuration file not found" << endl;
    return 1;
  };
  /* keep information about user */
  if(fill_user_spec(&user,auth,delegated_cred,handle) != 0) return 1;
  string pluginpath="/";
  bool right_group = true;
  for(;;) {
    string rest;
    string command=config_read_line(cfile,rest);
    if(command.length() == 0) break; /* EOF */
    if(command == "encryption") {  /* is encryption allowed ? */
      string value=config_next_arg(rest);
      if(value == "yes") {
        heavy_encryption=true;
      } else if(value == "no") {
        heavy_encryption=false;
      } else {
        user.Groups.clear();
        nodes.clear();
        olog<<"Error: improper attribute for encryption command: "<<value<<endl;
        return 1;
      };
    }
    if(command == "allowunknown") { /* should user be present in grid-mapfile ? */
      string value=config_next_arg(rest);
      if(value == "yes") {
        user.gridmap=true;
      } else if(value == "no") {
        if(!user.gridmap) {
          olog << "Error: unknown (non-gridmap) user is not allowed" << endl;
          return 1;
        };
      } else {
        user.Groups.clear();
        nodes.clear();
        olog<<"Error: improper attribute for encryption command: "<<value<<endl;        return 1;
      };
    }
    else if(command == "group") { /* definition of group of users based on DN */
      if(!right_group) {
        for(;;) {
          command=config_read_line(cfile,rest);
          if(command.length() == 0) break; /* eof - bad */
          if(command == "end") break;
        };
        continue;
      };
      string group_name=config_next_arg(rest);
      int decision = AAA_NO_MATCH;
      for(;;) {
        string rest=config_read_line(cfile);
        if(rest.length() == 0) break; /* eof - bad */
        if(rest == "end") break;
        if(decision == AAA_NO_MATCH) {
          decision = user.user.evaluate(rest.c_str());
          if(decision == AAA_FAILURE) decision=AAA_NO_MATCH;
        };
      };
      if(decision == AAA_POSITIVE_MATCH) user.Groups.push_back(group_name);
/*
      for(;;) {
        string user_subject=config_read_line(cfile,rest);
        if(user_subject.length() == 0) break; // eof - bad 
        if(user_subject == "end") break;
        if(user_subject == "file") {
          config_strip(rest);
          if(check_gridmap(user.user.DN(),NULL,rest.c_str())) {
            user.Groups.push_back(group_name);
          };
        }
        else {
          user_subject+=rest;
          if(user_subject == user.user.DN()) {
            user.Groups.push_back(group_name);
          };
        };
      };
*/
    }
    else if(command == "groupcfg") {  /* next commands only for these groups */
      if(rest.find_first_not_of(" \t") == string::npos) {
        right_group=true; continue;
      };
      right_group=false;
      for(;;) {
        string group_name=config_next_arg(rest);
        if(group_name.length() == 0) break;
        for(list<string>::iterator i=user.Groups.begin();
                                     i!=user.Groups.end();++i) {
          if((*i) == group_name) {
            right_group=true; break;
          };
        };
        if(right_group) break;
      };
    }
    else if(command == "pluginpath") {
      if(!right_group) continue;
      pluginpath=config_next_arg(rest);
      if(pluginpath.length() == 0) pluginpath="/";
    }
    else if(command == "plugin") {
      if(!right_group) {
        for(;;) {
          command=config_read_line(cfile,rest);
          if(command.length() == 0) break; /* eof - bad */
          if(command == "end") break;
        };
        continue;
      };
      string dir = config_next_arg(rest);
      string plugin = config_next_arg(rest);
      if(plugin.length() == 0) {
        olog << "Warning: can't parse configuration line" << endl;
        continue;
      };
      dir=subst_user_spec(dir,&user);
      if(canonical_dir(dir,false) != 0) {
        olog << "Warning: bad directory in plugin directive" << endl;
        continue;
      };
      /* look if path is not alredy registered */
      bool already_have_this_path=false;
      for(list<FileNode>::iterator i=nodes.begin();i!=nodes.end();++i) {
        if((*i) == dir) {
          already_have_this_path=true;
          break;
        };
      };
      if(already_have_this_path) {
        olog << "Already have directory:"<<dir<<endl;
        for(;;) {
          command=config_read_line(cfile,rest);
          if(command.length() == 0) break; /* eof - bad */
          if(command == "end") break;
        };
      }
      else {
        olog<<"Registering directory:"<<dir<<" with plugin: "<<plugin<<endl;
        plugin=pluginpath+'/'+plugin;
        FileNode node((char*)(dir.c_str()),(char*)(plugin.c_str)(),cfile,user);
        if(node.has_plugin()) { nodes.push_back(node); }
        else {
          olog << "Error: file node creation failed: " << dir << endl;
          for(;;) {
            command=config_read_line(cfile,rest);
            if(command.length() == 0) break; /* eof - bad */
            if(command == "end") break;
          };
        };
      };
    }
    else {
      olog << "Warning: unsupported configuration command" << endl;
    };
  };  
  config_close(cfile);
  if(!user.gridmap) {
    olog << "Error: unknown (non-gridmap) user is not allowed" << endl;
    return 1;
  };
  /* must be sorted to make sure we can always find right directory */
  nodes.sort(FileNode::compare);
  /* create dummy directories */
  int nn=nodes.size();
  for(list<FileNode>::iterator i=nodes.begin();i!=nodes.end();++i) {
    if(nn==0) break;
    string name = i->point;
    for(;remove_last_name(name);) {
      if(name.length() == 0) break;
      bool found = false;
      list<FileNode>::iterator ii=i;
      for(;;) {
        ++ii; if(ii == nodes.end()) break;
        if(name == ii->point) { found=true; break; };
      };
      if(!found) { /* add dummy dir */
        olog<<"Registering dummy directory:"<<name<<endl;
        nodes.push_back(FileNode(name.c_str()));
      }
      else {
        break;
      };
    };
    nn--;
  };
  opened_node=nodes.end();
  user.free();
  return 0;
}

FileRoot::FileRoot(void) {
  cur_dir="";
  opened_node=nodes.end();
  heavy_encryption=true;
};

int FileNode::readdir(const char* name,list<DirEntry> &dir_list,DirEntry::object_info_level mode) {
  if(plug) {
    return plug->readdir(remove_head_dir_c(name,point.length()),dir_list,mode);
  };
  return 0;
}

int FileNode::checkfile(string &name,DirEntry &info,DirEntry::object_info_level mode) {
// olog<<"FileNode: checkfile: "<<name<<endl;
  if(plug) {
    string dname=remove_head_dir_s(name,point.length());
    return plug->checkfile(dname,info,mode);
  };
  return 1;
}

int FileNode::checkdir(string &dirname) {
  if(plug) {
    string dname=remove_head_dir_s(dirname,point.length());
    if(plug->checkdir(dname) == 0) {
      dirname=point+'/'+dname;
      return 0;
    };
  };
  return 1;
}

int FileNode::makedir(string &dirname) {
  if(plug) {
    string dname=remove_head_dir_s(dirname,point.length());
    return plug->makedir(dname);
  };
  return 1;
}

int FileNode::removedir(string &dirname) {
  if(plug) {
    string dname=remove_head_dir_s(dirname,point.length());
    return plug->removedir(dname);
  };
  return 1;
}

int FileNode::removefile(string &name) {
  if(plug) {
    string dname=remove_head_dir_s(name,point.length());
    return plug->removefile(dname);
  };
  return 1;
}

int FileNode::open(const char* name,open_modes mode,unsigned long long int size) {
  if(plug) return plug->open(remove_head_dir_c(name,point.length()),mode,size);
  return 1;
}

int FileNode::close(bool eof) {
  if(plug) return plug->close(eof);
  return 1;
}

int FileNode::read(unsigned char *buf,unsigned long long int offset,unsigned long long *size) {
  if(plug) return plug->read(buf,offset,size);
  return 1;
}

int FileNode::write(unsigned char *buf,unsigned long long int offset,unsigned long long size) {
  if(plug) return plug->write(buf,offset,size);
  return 1;
}

