#ifndef __HTTPSD_H__
#define __HTTPSD_H__

#include <globus_io.h>
#include <list>
#include <string>

#include "../auth/auth.h"

class HTTP_Service;
class HTTP_Services;
class AuthEvaluator;

class HTTPS_Connector {
 private:
  static list<HTTPS_Connector*> cons;
  static pthread_mutex_t cons_lock;
  bool initialized;
  globus_io_handle_t h;
  HTTP_Services& services;
  list<AuthEvaluator*>& auths;
//  string subject;
  AuthUser user;
  static void accept_cb(void *callback_arg,globus_io_handle_t *handle,globus_result_t result);
 public:
  HTTPS_Connector(globus_io_handle_t *s,HTTP_Services& serv,list<AuthEvaluator*>& auths);
  ~HTTPS_Connector(void);
  operator bool(void) {
    return initialized;
  };
  size_t read(char* buf, size_t l);
  int write(const char* buf, size_t l);
  size_t readline(char* buf, size_t l);
//  bool add_service(const char *url,HTTP_Service *service);
//  bool remove_service(const char *url);
  void loop(void);
  static void identity(globus_io_handle_t* handle,const char* subject,gss_cred_id_t cred);
  const char* identity_subject(void) const { return user.DN(); };
  const char* identity_proxy(void) const { return user.proxy(); };
  AuthUser& identity(void) { return user; };
  list<AuthEvaluator*>& authorizations(void) { return auths; };
};

typedef enum {
  HTTP_OK = 200,
  HTTP_NOT_IMPLEMENTED = 501,
  HTTP_NOT_ALLOWED = 403,
  HTTP_NOT_FOUND = 404,
  HTTP_ERROR = 500
} HTTP_Error;

int skip_header(HTTPS_Connector &c,int &keep_alive);
int send_response_header(HTTPS_Connector &c,int keep_alive,int code,char* type,int size);
int send_file(const char* fname,HTTPS_Connector &c);
int stat_file(const char* fname,unsigned long long int &size);
int send_error_response(HTTPS_Connector &c,int keep_alive,int code,char* type,char* content);


class HTTP_Service {
 public:
  HTTP_Service(void);
  virtual ~HTTP_Service(void);
  virtual HTTP_Error get(int &keep_alive);
  virtual HTTP_Error put(int &keep_alive);
  virtual HTTP_Error post(int &keep_alive);
};

class HTTP_Services;

typedef HTTP_Service* (*service_creator)(HTTPS_Connector& c,const char* uri,void* arg);
typedef void* (*service_configurator)(istream& f,const char* uri);

class HTTP_Service_description {
 friend class HTTP_Services;
 private:
  string uri;
  service_creator creator;
  void* arg;
 public:
  HTTP_Service_description(const char* u,service_creator c,void* arg);
  bool match(const char* u);
  HTTP_Service* create(HTTPS_Connector& c);
};

class HTTP_Service_item {
 public:
  const HTTPS_Connector* con;
  string uri;
  HTTP_Service* serv;
  HTTP_Service_item(const HTTPS_Connector* c,const char* u,HTTP_Service* s):
     con(c),uri(u),serv(s) { };
  ~HTTP_Service_item(void) { };
};

class HTTP_Services {
 private:
  list<HTTP_Service_description> desc;
  list<HTTP_Service_item> serv;
  pthread_mutex_t lock;
 public:
  HTTP_Services(void);
  ~HTTP_Services(void);
  bool add(const char* u,service_creator c,void* arg);
  bool remove(const char* u,service_creator c);
  HTTP_Service* get(HTTPS_Connector& c,const char* u);
  void close(const HTTPS_Connector& c);
};

#endif /* __HTTPSD_H__ */
