/* write essential inforamtion about job started/finished */
#include "../std.h"
#include <fstream>
#include "../files/info_types.h"
#include "../files/info_files.h"
#include "../config/environment.h"
#include "log_time.h"
#include "job_log.h"

JobLog::JobLog(void):filename(""),url(""),cluster_name(""),proc(NULL),last_run(0) {
}

JobLog::JobLog(const char* fname):url(""),cluster_name(""),proc(NULL),last_run(0) {
  filename=fname;
}

void JobLog::SetOutput(const char* fname) {
  filename=fname;
}

bool JobLog::open_stream(ofstream &o) {
    o.open(filename.c_str(),ofstream::app);
    if(!o.is_open()) return false;
    o<<" ";
    LogTime::gmdatetime(o);
    return true;
}

bool JobLog::start_info(JobDescription &job,const JobUser &user) {
  if(filename.length()==0) return true;
    ofstream o;
    if(!open_stream(o)) return false;
    o<<"Started - job id: "<<job.get_id()<<", unix user: "<<job.get_uid()<<":"<<job.get_gid()<<", ";
    if(job.GetLocalDescription(user)) {
      JobLocalDescription *job_desc = job.get_local();
      string tmps;
      tmps=job_desc->jobname; make_escaped_string(tmps,'"');
      o<<"name: \""<<tmps<<"\", ";
      tmps=job_desc->DN; make_escaped_string(tmps,'"');
      o<<"owner: \""<<tmps<<"\", ";
      o<<"lrms: "<<job_desc->lrms<<", queue: "<<job_desc->queue;
    };
    o<<endl;
    o.close();
}

bool JobLog::finish_info(JobDescription &job,const JobUser &user) {
  if(filename.length()==0) return true;
    ofstream o;
    if(!open_stream(o)) return false;
    o<<"Finished - job id: "<<job.get_id()<<", unix user: "<<job.get_uid()<<":"<<job.get_gid()<<", ";
    string tmps;
    if(job.GetLocalDescription(user)) {
      JobLocalDescription *job_desc = job.get_local();
      tmps=job_desc->jobname; make_escaped_string(tmps,'"');
      o<<"name: \""<<tmps<<"\", ";
      tmps=job_desc->DN; make_escaped_string(tmps,'"');
      o<<"owner: \""<<tmps<<"\", ";
      o<<"lrms: "<<job_desc->lrms<<", queue: "<<job_desc->queue;
      if(job_desc->localid.length() >0) o<<", lrmsid: "<<job_desc->localid;
    };
    tmps = job.GetFailure();
    if(tmps.length()) {
      for(int i=0;;){i=tmps.find('\n',i);if(i==string::npos)break;tmps[i]='.';};
      make_escaped_string(tmps,'"');
      o<<", failure: \""<<tmps<<"\"";
    };
    o<<endl;
    o.close();
} 


bool JobLog::read_info(fstream &i,bool &processed,bool &jobstart,struct tm &t,JobId &jobid,JobLocalDescription &job_desc,string &failure) {
  processed=false;
  if(!i.is_open()) return false;
  char line[4096];
  streampos start_p=i.tellp();
  istream_readline(i,line,sizeof(line));
  streampos end_p=i.tellp();
  if((line[0] == 0) || (line[0] == '*')) { processed=true; return true; };
  char* p = line;
  if((*p) == ' ') p++;
  // struct tm t;
  /* read time */
  if(sscanf(p,"%u-%u-%u %u:%u:%u ",
       &t.tm_mday,&t.tm_mon,&t.tm_year,&t.tm_hour,&t.tm_min,&t.tm_sec) != 6) {
    return false;
  };
  /* skip time */
  for(;(*p) && ((*p)==' ');p++); if(!(*p)) return false;
  for(;(*p) && ((*p)!=' ');p++); if(!(*p)) return false;
  for(;(*p) && ((*p)==' ');p++); if(!(*p)) return false;
  for(;(*p) && ((*p)!=' ');p++); if(!(*p)) return false;
  for(;(*p) && ((*p)==' ');p++); if(!(*p)) return false;
  // bool jobstart;
  if(strncmp("Finished - ",p,11) == 0) {
    jobstart=false; p+=11;
  } else if(strncmp("Started - ",p,10) == 0) {
    jobstart=true; p+=10;
  } else {
    return false;
  };
  /* read values */
  char* name;
  char* value;
  char* pp;
  for(;;) {
    for(;(*p) && ((*p)==' ');p++); if(!(*p)) break;
    if((pp=strchr(p,':')) == NULL) break;
    name=p; (*pp)=0; pp++;
    for(;(*pp) && ((*pp)==' ');pp++);
    value=pp;
    if((*value) == '"') {
      value++;
      pp=make_unescaped_string(value,'"');
      for(;(*pp) && ((*pp) != ',');pp++);
      if((*pp)) pp++;
    } else {
      for(;(*pp) && ((*pp) != ',');pp++);
      if((*pp)) { (*pp)=0; pp++; };
    };
    p=pp;
    /* use name:value pair */
    if(strcasecmp("job id",name) == 0) {
      jobid=value;
    } else if(strcasecmp("name",name) == 0) {
      job_desc.jobname=value;
    } else if(strcasecmp("unix user",name) == 0) {

    } else if(strcasecmp("owner",name) == 0) {
      job_desc.DN=value;
    } else if(strcasecmp("lrms",name) == 0) {
      job_desc.lrms=value;
    } else if(strcasecmp("queue",name) == 0) {
      job_desc.queue=value;
    } else if(strcasecmp("lrmsid",name) == 0) {
      job_desc.localid=value;
    } else if(strcasecmp("failure",name) == 0) {
      failure=value;
    } else {

    };
  };
  i.seekp(start_p); i<<"*"; i.seekp(end_p);
  return true;
}

bool JobLog::RunReporter(JobUsers &users) {
  if(!is_reporting()) return true;
  if(proc != NULL) {
    if(proc->get_exit_code() == -1) return true; /* running */
    RunParallel::release(proc);
    proc=NULL;
  };
  if(time(NULL) < (last_run+3600)) return true; // once per hour
  last_run=time(NULL);
  if(users.size() <= 0) return true; // no users to report
  char** args = (char**)malloc(sizeof(char*)*(users.size()+6)); 
  if(args == NULL) return false;
  string cmd = nordugrid_bin_loc+"/logger";
  int argc; args[argc++]=(char*)cmd.c_str();
  args[argc++]=(char*)"-u"; args[argc++]=(char*)url.c_str();
  if(cluster_name.length()) {
    args[argc++]=(char*)"-c"; args[argc++]=(char*)cluster_name.c_str();
  };
  for(JobUsers::iterator i = users.begin();i != users.end();++i) {
    args[argc++]=(char*)(i->ControlDir().c_str());
  };
  args[argc]=NULL;
  JobUser user(getuid());
  return RunParallel::run(user,NULL,args,&proc,false,false);
}

bool JobLog::SetReporter(const char* destination,const char* name) {
  if(destination) url=destination;
  if(name) cluster_name=name;
  return true;
}

bool JobLog::make_file(const JobDescription &job,JobUser &user) {
  if(!is_reporting()) return true;
  return job_log_make_file(job,user);
}

JobLog::~JobLog(void) {
  if(proc != NULL) {
    if(proc->get_exit_code() == -1) proc->kill();
    RunParallel::release(proc);
    proc=NULL;
  };
}

