#ifndef GRID_MANAGER_RUN_H
#define GRID_MANAGER_RUN_H

#include "../std.h"
#include "../jobs/users.h"
#include "../jobs/states.h"
#include <sys/resource.h>
#include <sys/wait.h>
#include <globus_common.h>

extern char** environ;

class RunParallel;
class RunElementPointer;

class RunElement {
 friend class RunParallel;
 friend class RunElementPointer;
 private:
  int pid;
  int exit_code;
  bool released;
  RunElement* next;
  RunElement* next_u;
 public:
  void kill(int signum = SIGTERM) {
    int tmp_pid=pid; if(tmp_pid != -1) ::kill(tmp_pid,signum);
  };
  int get_exit_code(void) { return exit_code; };
  RunElement() { pid=0; exit_code=-1; released=false; next=NULL; next_u=NULL; };
  RunElement(const RunElement &re) { 
    pid=re.pid; exit_code=re.exit_code; released=re.released;
    next=NULL; next_u=NULL;
  };
  ~RunElement() { };
};

class RunElementPointer {
 friend class RunParallel;
 private:
  RunElement* prev;
  RunElement* cur;
 public:
  RunElementPointer(void) { cur=NULL; prev=NULL; };
  RunElementPointer(RunElement* re) { cur=re; prev=NULL; };
  ~RunElementPointer(void) { };
  bool is_end(void) { return (cur==NULL); };
  void go_next(void) { if(cur) { prev=cur; cur=cur->next; }; };
  void go_next_u(void) { if(cur) { prev=cur; cur=cur->next_u; };
  };
  void operator=(RunElement* re) { prev=NULL; cur=re; };
  RunElement* operator->() { return cur; };
};

class RunParallel {
 private:
  static bool hup_detected;
  bool initialized;
  static RunElement dumb_re;
  static bool in_signal;
  static globus_cond_t  *cond;
  static struct sigaction old_sig_chld; 
  static struct sigaction old_sig_hup; 
  static struct sigaction old_sig_term; 
  static bool old_sig_chld_inited; 
  static bool old_sig_hup_inited; 
  static bool old_sig_term_inited; 
  static void sig_hup(int signum,siginfo_t *info,void* arg);
  static void sig_chld(int signum,siginfo_t *info,void* arg);
  static void sig_term(int signum,siginfo_t *info,void* arg);
  static RunElement* begin;
  static RunElement* begin_u;
  bool init(void);
  void deinit(void);
 public:
  static bool run(JobUser& user,const char* jobid,char *const args[],RunElement**,bool su = true,bool job_proxy = true);
  static bool run(JobUser& user,const JobDescription& desc,char *const args[],RunElement**,bool su = true);
  static void release(RunElement* re);
  static RunElement* add_handled(RunElement* re);
  bool is_initialized(void) { return initialized; };
  bool was_hup(void);
  RunParallel(globus_cond_t *cond);
  ~RunParallel();
};

#endif // GRID_MANAGER_RUN_H
