#include "../std.h"
#include <globus_common.h>
#include <globus_gass_copy.h>
#include <globus_ftp_client_plugin.h>
#include <iostream>
#include "../misc/log_time.h"
#include "../misc/url_options.h"
#include "../cache/cache.h"
#include "check.h"
#include "ftp_preallocate_plugin.h"
#include "cache_download.h"

#define GLOBUS_ERROR GLOBUS_ERROR_NO_INFO

typedef struct {
  void *arg;
  globus_gass_copy_callback_t cb;
  globus_gass_copy_handle_t *handle;
/* the pointers below in this sctructure should be freed at the end */
  const char* s_url;
  globus_gass_copy_attr_t *s_attr;
  const char* d_url;
  globus_gass_copy_attr_t *d_attr;
  const char* cache_url;
  const char* cache_path;
  const char* cache_data_path;
  const char* cache_link_path;
  uid_t cache_uid;
  gid_t cache_gid;
  const char* id;
  globus_mutex_t done_mutex;
  globus_cond_t done_cond;
  int done_flag;
} cache_args_t;

/* called if out of disk space */
int preallocate_callback(globus_ftp_client_plugin_t *plugin,const char* path,off_t size,void* arg) {
  cache_args_t* cargs = (cache_args_t*)arg;
  olog<<"Cache cleaning requested: "<<cargs->cache_path<<", "<<cargs->cache_data_path<<", "<<size<<" bytes"<<endl;
  unsigned long long int freed=cache_clean(cargs->cache_path,cargs->cache_data_path,cargs->cache_uid,cargs->cache_gid,size);
  olog<<"Cache cleaned: "<<cargs->cache_path<<", "<<cargs->cache_path<<", "<<freed<<" bytes"<<endl;
  if(freed < size) return -1;
  return 0;
}

bool make_link(const char *file_url,const char *link_url,const char *cache_path,const char *cache_data_path,const char *cache_link_path)  {
  bool res_succ = true;
  const char* s_file_name = get_url_path(file_url);
  if(s_file_name == NULL) {
    olog<<"Failed extracting path from "<<file_url<<endl;
    res_succ=false;
  }
  else {
    s_file_name+=strlen(cache_data_path);
    char* s_file = (char*)malloc(strlen(cache_link_path)+
                                        strlen(s_file_name)+1);
    if(s_file == NULL) {
      res_succ=false;
      olog<<"Failed allocating memory for link name"<<endl;
    }
    else {
      strcpy(s_file,cache_link_path);
      strcat(s_file,s_file_name);
      const char* d_link = get_url_path(link_url);
      if(d_link == NULL) {
        res_succ=false;
        olog<<"Failed extracting path from "<<link_url<<endl;
      }
      else {
        if(symlink(s_file,d_link) == -1) {
          perror("symlink");
          olog<<"Failed to make a link "<<d_link<<" to file "<<s_file<<endl;
          res_succ=false;
        };
        // chmod(s_file,S_IRUSR,S_IWUSR);
      };
    };
    free(s_file);
  };
  return res_succ;
}

void cache_callback(void *arg,globus_gass_copy_handle_t *handle,globus_object_t *error) {
  cache_args_t *cargs = (cache_args_t*)arg;
  globus_mutex_lock(&(cargs->done_mutex));
  if(error == GLOBUS_SUCCESS) { cargs->done_flag=1; }
  else { cargs->done_flag=2; };
  globus_cond_signal(&(cargs->done_cond));
  globus_mutex_unlock(&(cargs->done_mutex));
}

/* TODO - check if all 'return' are ok */
void* cache_load(void* args) {
  globus_result_t res;
  /* read status of cache file and claim download access */
  cache_args_t* cargs = (cache_args_t*)args; /* this should be freed at end */

  cache_download_handler cdh;
  globus_ftp_client_plugin_t pplg;
  bool preallocate_plugin = false;

  globus_mutex_init(&(cargs->done_mutex),GLOBUS_NULL);
  globus_cond_init(&(cargs->done_cond),GLOBUS_NULL);
  cargs->done_flag=0;
  switch(cache_download_url_start(
              cargs->cache_path,cargs->cache_data_path,
              cargs->cache_uid,cargs->cache_gid,
              cargs->cache_url,cargs->id,cdh)) {
    case 1: { // error 
      olog<<"Error while locking file in cache"<<endl;
      (*(cargs->cb))(cargs->arg,cargs->handle,GLOBUS_ERROR_NO_INFO);
    }; break;
    case 2: {  // already have it
      char* file_url=(char*)malloc(cdh.name().length()+8);
      if(file_url == NULL) {
        olog<<"Failed to allocate memory"<<endl;
        (*(cargs->cb))(cargs->arg,cargs->handle,GLOBUS_ERROR_NO_INFO);
      }
      else {
        strcpy(file_url,"file://");
        strcat(file_url,cdh.name().c_str());
        // need to check access permissions
        res=globus_gass_copy_url_access_check(
                           (char*)(cargs->s_url),cargs->s_attr);
        if(res != GLOBUS_SUCCESS) {
          olog<<"Permission check failed"<<endl;
          (*(cargs->cb))(cargs->arg,cargs->handle,GLOBUS_ERROR_NO_INFO);
        }
        else {
          if(make_link(file_url,cargs->d_url,cargs->cache_path,
                            cargs->cache_data_path,cargs->cache_link_path)) {
            (*(cargs->cb))(cargs->arg,cargs->handle,GLOBUS_SUCCESS);
          }
          else {
            olog<<"Failed to make soft link"<<endl;
            (*(cargs->cb))(cargs->arg,cargs->handle,GLOBUS_ERROR_NO_INFO);
          };
        };
      };
      if(file_url) free(file_url);
    }; break;
    case 0: {  // have to download
      char* file_url=(char*)malloc(cdh.name().length()+8);
      if(file_url == NULL) {
        olog<<"Failed to allocate memory"<<endl;
        cache_download_url_end(cdh,false);
        (*(cargs->cb))(cargs->arg,cargs->handle,GLOBUS_ERROR_NO_INFO);
      }
      else {
        strcpy(file_url,"file://");
        strcat(file_url,cdh.name().c_str());
        // add preallocate plugin 
        if(globus_ftp_client_preallocate_plugin_init(&pplg,
                          file_url,&preallocate_callback,cargs
                                                ) == GLOBUS_SUCCESS) {
          if(globus_ftp_client_handle_add_plugin(
                     &(cargs->handle->ftp_source_handle),
                                    &pplg) != GLOBUS_SUCCESS) {
            olog<<"cache_load: adding preallocate plugin FAILED"<<endl;
            globus_ftp_client_preallocate_plugin_destroy(&pplg);
          }
          else { preallocate_plugin=true; };
        };
        // using *_check  to handle link: urls
        if((res=globus_gass_copy_register_url_to_url_check(
                   cargs->handle,
                   (char*)(cargs->s_url),cargs->s_attr,
                   file_url,cargs->d_attr,
                   &cache_callback,cargs)) != GLOBUS_SUCCESS) {
          if(preallocate_plugin) {
            globus_ftp_client_handle_remove_plugin(
                   &(cargs->handle->ftp_source_handle),&pplg);
            globus_ftp_client_preallocate_plugin_destroy(&pplg);
          };
          cache_download_url_end(cdh,false);
          (*(cargs->cb))(cargs->arg,cargs->handle,globus_error_get(res));
        }
        else {
          bool res_succ;
          globus_mutex_lock(&(cargs->done_mutex));
          while(cargs->done_flag == 0) {
            globus_cond_wait(&(cargs->done_cond),&(cargs->done_mutex));
            res_succ=(cargs->done_flag == 1);
          };
          globus_mutex_unlock(&(cargs->done_mutex));
          if(preallocate_plugin) {
            globus_ftp_client_handle_remove_plugin(
                   &(cargs->handle->ftp_source_handle),&pplg);
            globus_ftp_client_preallocate_plugin_destroy(&pplg);
          };
          if(res_succ) { 
            cache_download_url_end(cdh,true);
            if(make_link(file_url,cargs->d_url,cargs->cache_path,
                     cargs->cache_data_path,cargs->cache_link_path)) {
              (*(cargs->cb))(cargs->arg,cargs->handle,GLOBUS_SUCCESS);
            }
            else {
              (*(cargs->cb))(cargs->arg,cargs->handle,GLOBUS_ERROR_NO_INFO);
            };
          }
          else {
            cache_download_url_end(cdh,false);
            (*(cargs->cb))(cargs->arg,cargs->handle,GLOBUS_ERROR_NO_INFO);
          };
        };
      };
      if(file_url) free(file_url);
    }; break;
    default: {
      (*(cargs->cb))(cargs->arg,cargs->handle,GLOBUS_ERROR_NO_INFO);
    };
  };
  /* free memory */
  globus_mutex_destroy(&(cargs->done_mutex));
  globus_cond_destroy(&(cargs->done_cond));
  if(cargs->s_url)           free((void*)(cargs->s_url));
  if(cargs->d_url)           free((void*)(cargs->d_url));
  if(cargs->cache_url)       free((void*)(cargs->cache_url));
  if(cargs->cache_path)      free((void*)(cargs->cache_path));
  if(cargs->cache_data_path) free((void*)(cargs->cache_data_path));
  if(cargs->cache_link_path) free((void*)(cargs->cache_link_path));
  if(cargs->id)              free((void*)(cargs->id));
  if(cargs->s_attr)          free((void*)(cargs->s_attr));
  if(cargs->d_attr)          free((void*)(cargs->d_attr));
  delete cargs;
  return NULL;
}

/*
  cache_url - url used to distinguish file in cache
  user - used for cache attributes 
*/
globus_result_t globus_gass_copy_register_url_to_url_cache(
    globus_gass_copy_handle_t *a_handle,
    const char* s_url,globus_gass_copy_attr_t *s_attr,
    const char* d_url,globus_gass_copy_attr_t *d_attr,
    const char* cache_url,const char* cache_path,
    const char* cache_data_path,const char* id,
    uid_t cache_uid,gid_t cache_gid,const char* cache_link_path,
    globus_gass_copy_callback_t cb,void* arg) {
  globus_result_t res;

  if(cache_link_path==NULL) cache_link_path=cache_data_path;
  /* destination must be local */
  if(strncmp(d_url,"file://",7) != 0) {
    olog<<"destination url must be local: "<<d_url<<endl;
    return GLOBUS_ERROR;
  };
  /* cache entry should be already initialized and claimed */    
  /* start new thread, because blocking function will be used and
     callback can deadlock too */
  globus_thread_t cache_thread;
  /* fill arguments for new thread */

  cache_args_t* args = new cache_args_t;
  if(args == NULL) {
    olog<<"Memory allocation error"<<endl; return GLOBUS_ERROR;
  };
  args->cb=cb;
  args->arg=arg;
  args->handle=a_handle;
  args->s_url=strdup(s_url);
  args->s_attr=NULL;
  if(s_attr) args->s_attr=
      (globus_gass_copy_attr_t*)malloc(sizeof(globus_gass_copy_attr_t));
  args->d_url=strdup(d_url);
  args->d_attr=NULL;
  if(d_attr) args->d_attr=
      (globus_gass_copy_attr_t*)malloc(sizeof(globus_gass_copy_attr_t));
  args->cache_url=strdup(cache_url);
  args->cache_path=strdup(cache_path);
  args->cache_data_path=strdup(cache_data_path);
  args->cache_link_path=strdup(cache_link_path);
  args->cache_uid=cache_uid;
  args->cache_gid=cache_gid;
  args->id=strdup(id);
  if((args->s_url == NULL) || (args->d_url == NULL) ||
     ((s_attr != NULL) && (args->s_attr == NULL)) ||
     ((d_attr != NULL) && (args->d_attr == NULL)) ||
     (args->cache_url == NULL) || (args->cache_path == NULL) || 
     (args->cache_data_path == NULL) ||
     (args->cache_link_path == NULL) || (args->id == NULL)) {
    if(args->s_url) free((void*)(args->s_url));
    if(args->s_attr) free(args->s_attr);
    if(args->d_url) free((void*)(args->d_url));
    if(args->d_attr) free(args->d_attr);
    if(args->cache_url) free((void*)(args->cache_url));
    if(args->cache_path) free((void*)(args->cache_path));
    if(args->cache_data_path) free((void*)(args->cache_data_path));
    if(args->cache_link_path) free((void*)(args->cache_link_path));
    if(args->id) free((void*)(args->id));
    delete args;
    olog<<"Memory allocation error"<<endl; return GLOBUS_ERROR;
  };
  if(s_attr) memcpy(args->s_attr,s_attr,sizeof(globus_gass_copy_attr_t));
  if(d_attr) memcpy(args->d_attr,d_attr,sizeof(globus_gass_copy_attr_t));
  if(globus_thread_create(&cache_thread,NULL,&cache_load,args) != 0) {
    olog<<"Failed to start new thread"<<endl;
    if(args->s_url)   free((void*)(args->s_url));
    if(args->s_attr)  free(args->s_attr);
    if(args->d_url)   free((void*)(args->d_url));
    if(args->d_attr)  free(args->d_attr);
    if(args->cache_url)  free((void*)(args->cache_url));
    if(args->cache_path) free((void*)(args->cache_path));
    if(args->cache_data_path) free((void*)(args->cache_data_path));
    if(args->cache_link_path) free((void*)(args->cache_link_path));
    if(args->id) free((void*)(args->id));
    delete args;
    return GLOBUS_ERROR;
  };
  return GLOBUS_SUCCESS;
}
