#include "../std.h"
#include <iostream>
#include <globus_ftp_client.h>
#include <globus_ftp_client_plugin.h>
#include <globus_ftp_client_debug_plugin.h>
#include <globus_gass_copy.h>

#include "cache.h"
#include "cache_download.h"
#include "ftp_preallocate_plugin.h"

globus_mutex_t download_lock;
globus_cond_t  download_cond;
int download_done = 0;

void download_callback(void * arg,globus_gass_copy_handle_t * handle,globus_object_t * error) {
  globus_mutex_lock(&(download_lock));
  if(error != GLOBUS_SUCCESS) {
    cerr<<"callback error: "<<globus_object_printable_to_string(error)<<endl;
cerr<<"setting download done to -1"<<endl;
    download_done=-1;
  }
  else {
cerr<<"setting download done to 1"<<endl;
    cerr<<"No errors"<<endl;
    download_done=1;
  };
  globus_cond_signal(&download_cond);
  globus_mutex_unlock(&(download_lock));
}

int main(int argc,char** argv) {
  int res;
  globus_result_t gres;
  const char* s_url="gsiftp://grid.uio.no/scratch/gdmp/test";
  const char* d_url="file:///tmp/test";
  const char* cache_url=strdup(s_url);
  const char* cache_path="/tmp/cache";
  const char* id="test_job";

/*
  unsigned long long int size = 100;
  size=cache_clean(cache_path,0,0,size);
  cerr<<size<<" bytes freed"<<endl;
  exit(0);
*/
/*
  int h = open("/tmp/a.file",O_CREAT | O_WRONLY);
  off_t o = 2000000000;
  if(h==-1) exit(1);
  if(lseek(h,o,SEEK_SET) != o) perror("lseek");
  if(ftruncate(h,o) == -1) perror("ftruncate");
  char c = 0xFF;
  off_t i;
  for(i=0;i<o;i+=4096) {
    if(lseek(h,i,SEEK_SET) != i) perror("lseek #2");
    if(write(h,&c,1) != 1) perror("write");
  };
  close(h);
  exit(0);
*/

  res=cache_find_url(cache_path,0,0,s_url,id);
  cout<<"Result: "<<res<<endl;
  if(res != 0)  return 1;

cerr<<"--- activate"<<endl;
  globus_module_activate(GLOBUS_COMMON_MODULE);
  globus_module_activate(GLOBUS_FTP_CLIENT_MODULE);
  globus_module_activate(GLOBUS_GASS_COPY_MODULE);
  globus_mutex_init(&download_lock,GLOBUS_NULL);
  globus_cond_init(&download_cond,GLOBUS_NULL);

  globus_gass_copy_handle_t handle;
  globus_gass_copy_attr_t s_attr;
  globus_gass_copy_attr_t d_attr;
  globus_gass_copy_handle_init(&handle,GLOBUS_NULL);
  globus_gass_copy_attr_init(&s_attr);
  globus_gass_copy_attr_init(&d_attr);


cerr<<"--- gass register "<<endl;
  gres=globus_gass_copy_register_url_to_url_cache(
    &handle,s_url,&s_attr,d_url,&d_attr,
    cache_url,cache_path,id,0,0,NULL,
    download_callback,NULL);
  if(gres != GLOBUS_SUCCESS) {
    cerr<<"Error: "<<globus_object_printable_to_string(globus_error_get(gres))<<endl;
    globus_module_deactivate(GLOBUS_GASS_COPY_MODULE);
    globus_module_deactivate(GLOBUS_COMMON_MODULE);
    return 1;
  };

  globus_mutex_lock(&(download_lock));
  while(!(download_done)) {
    cerr<<"Waiting for download"<<endl;
    globus_cond_wait(&(download_cond),&(download_lock));
  };
  globus_mutex_unlock(&(download_lock));

  if(download_done==-1) {
cerr<<"release and delete"<<endl;
    res=cache_release_url("/tmp/cache",0,0,"test_job",true);
  }
  else {
cerr<<"release and not delete"<<endl;
    res=cache_release_url("/tmp/cache",0,0,"test_job",false);
  };
  cout<<"Result: "<<res<<endl;

cerr<<"--- destroy handle"<<endl;
  globus_gass_copy_handle_destroy(&handle);

  globus_cond_destroy(&download_cond);
  globus_mutex_destroy(&download_lock);

cerr<<"--- deactivate"<<endl;
  globus_module_deactivate(GLOBUS_GASS_COPY_MODULE);
  globus_module_deactivate(GLOBUS_FTP_CLIENT_MODULE);
  globus_module_deactivate(GLOBUS_COMMON_MODULE);
  return 0;
}
