#include "../std.h"
#include "../config/config_map.h"
#include "replica_utils.h"
#include "../misc/log_time.h"
#include "convert_url.h"

/* Get the url for upload. If url is RC, also extract ldap url */
bool convert_url(string &url,FileDataRC &fd,UrlMap &url_map,bool existing,bool non_existing,RCFile *ex_file) {
cerr<<"convert_url: url"<<fd.lfn<<endl;
  /* determine what kind of url it is */
  /* const char *s = url.c_str(); */
  const char *s = fd.lfn.c_str();
  url=fd.lfn;
  olog << "convert url: " << url << "  " << (long int)fd.locs << endl;
  if(!strncasecmp(s,rc_url_head,strlen(rc_url_head))) {
    if(fd.locs == NULL) {
      olog << "Making new locs" << endl;
      /* this url starts from rc:// - extract ldap:// from it */
      fd.rc_url.resize(0);
      if(!extract_RC_from_url(url,fd.rc_url)) {
        olog<<"URL part to contact Replica Catalog is missing from "<<url<<endl;
        return false;
      };
      /* extract host and filename from url */
      s=url.c_str()+strlen(rc_url_head);
      const char* s_e = strchr(s,'/'); if(s_e == NULL) s_e=s+strlen(s);
      string host(s,s_e-s);
      olog<<"Location host: "<<host<<endl;
      RCManager rc(fd.rc_url,"","");
      if(!(rc.is_open())) {
        olog<<"Failed accessing Replica Catalog collection: "<<fd.rc_url<<endl;
        return false;
      };
      fd.locs = new list<RCLocation>;
      const char* filename = NULL;
      if((*s_e) != 0) filename=(s_e+1);
      if(filename) {
        olog<<"Location filename: "<<filename<<endl;
        if(ex_file) {
          RCFile t_file(filename,0,0,0);
          if(rc.GetFile(t_file)) (*ex_file)=t_file;
        };
      };
      if(existing && non_existing) {
        if(!(rc.GetLocations(host,*fd.locs))) {
          olog << "Failed resolving location: " << host << endl;
          delete fd.locs;
          return false;
        };
      }
      else if(existing) {
        if(filename == NULL) {
          olog << "No filename is given in " << url << endl;
          delete fd.locs;
          return false;
        };
        if(!(rc.GetLocations(host,*fd.locs,filename))) {
          olog << "Failed resolving location: " << host << endl;
          delete fd.locs;
          return false;
        };
      }
      else if(non_existing) {
        if(!(rc.GetLocations(host,*fd.locs))) {
          olog << "Failed resolving location: " << host << endl;
          delete fd.locs;
          return false;
        };
        if(filename != NULL) {
          list<RCLocation> t_locs;
          if(rc.GetLocations(host,t_locs,filename)) {
            for(list<RCLocation>::iterator i1=fd.locs->begin();i1!=fd.locs->end();) {
              bool erased_location = false;
              for(list<RCLocation>::iterator i2=t_locs.begin();i2!=t_locs.end();++i2) {
                if(i1->url == i2->url) {
                  i1=fd.locs->erase(i1);
                  t_locs.erase(i2);
                  erased_location=true;
                  break;
                };
              };
              if(!erased_location) ++i1;
            };
          };
        };
      }
      else { /* nothing requested */
        return true;
      };
      if(fd.locs->size() == 0) {
        olog << "No suitable locations found among: " << host << endl;
        delete fd.locs;
        return false;
      };
      for(list<RCLocation>::iterator i=fd.locs->begin();i!=fd.locs->end();++i) {
        if((*s_e) != 0) i->url+=(s_e+1);
olog<<"location(1): "<< i->url << endl;
      };
      /* sort locations - put local first */
      for(list<RCLocation>::iterator i=fd.locs->begin();i!=fd.locs->end();) {
        if(url_map.local(i->url)) {
          fd.locs->push_front(*i);
          i=fd.locs->erase(i);
          continue;
        };
        ++i;
      };
      for(list<RCLocation>::iterator i=fd.locs->begin();i!=fd.locs->end();++i) {
        olog<<"location(2): "<< i->url << endl;
      };
    }
    else { olog << "Using old locations" << endl; };
    if(fd.try_url >= fd.locs->size()) fd.try_url = 0;
    olog << "trying url " << fd.try_url << endl;
    int n=0; list<RCLocation>::iterator i=fd.locs->begin();
    for(;i!=fd.locs->end();++i) {
      if(n == fd.try_url) {
        olog << "trying url " << i->url << endl;
        fd.try_url++; url=i->url; return true;
      }; n++;
    };
    return false;
  };
  return true;
};

