#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <globus_common.h>
#include <globus_ftp_client.h>
#include <globus_ftp_client_debug_plugin.h>
#include <globus_gass_copy.h>
#include <globus_object.h>


globus_mutex_t upload_lock;
globus_cond_t  upload_cond;
int upload_done=0;

void upload_finished_callback(void * arg,globus_gass_copy_handle_t * handle,globus_object_t * error) {

fprintf(stderr,"Transfer callback\n");
  if(error != GLOBUS_SUCCESS) {
fprintf(stderr,"Transfer failed\n");
fprintf(stderr,"Error is: %s\n",globus_object_printable_to_string(error));
  }
  else {
  };
  globus_mutex_lock(&(upload_lock));
  upload_done=1;
  globus_cond_signal(&upload_cond);
  globus_mutex_unlock(&(upload_lock));
}

int main(int argc,char** argv) {
  int n_threads;
  
  char* from_url;
  char* to_url;


  globus_gass_copy_handle_t handle;
  globus_ftp_client_operationattr_t ftp_attr;
  globus_gass_copy_attr_t o_attr;
  globus_gass_copy_handleattr_t      gass_copy_handleattr;
  globus_ftp_client_handleattr_t     ftp_handleattr;
  globus_ftp_client_plugin_t         debug_plugin;
  globus_ftp_control_parallelism_t ftp_paral;
  globus_result_t error;

  from_url = argv[1];
  to_url = argv[2];
  n_threads = 1;

  globus_module_activate(GLOBUS_GASS_COPY_MODULE);
  globus_module_activate(GLOBUS_FTP_CLIENT_DEBUG_PLUGIN_MODULE);

  globus_gass_copy_handleattr_init(&gass_copy_handleattr);
  globus_ftp_client_handleattr_init(&ftp_handleattr);
  globus_ftp_client_debug_plugin_init(&debug_plugin,stderr,"debug");
  globus_ftp_client_handleattr_add_plugin(&ftp_handleattr,&debug_plugin);
  globus_gass_copy_handleattr_set_ftp_attr(&gass_copy_handleattr,&ftp_handleattr);

  if(globus_gass_copy_handle_init(&handle,&gass_copy_handleattr) != GLOBUS_SUCCESS) {
    fprintf(stderr,"Failed to init globus handle"); return 2;
  };
  globus_ftp_client_operationattr_init(&ftp_attr);
  if(n_threads>1) {
    ftp_paral.mode=GLOBUS_FTP_CONTROL_PARALLELISM_FIXED;
    ftp_paral.fixed.size=n_threads;
  }
  else {
    ftp_paral.mode=GLOBUS_FTP_CONTROL_PARALLELISM_NONE;
/* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  because parallelism is set to NONE the fixes.size (or base.size)
  should have no sense. Nevertheless if set to 
    =0 - data connection is never setup
    >0 - things work as expected
    <0 - segmentation fault
  This all happens ONLY if MODE E is used.
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */
    ftp_paral.base.size=-1;
    globus_ftp_client_operationattr_set_parallelism(&ftp_attr,&ftp_paral);
  };
    globus_ftp_client_operationattr_set_striped(&ftp_attr,GLOBUS_FALSE);
    globus_ftp_client_operationattr_set_type(&ftp_attr,GLOBUS_FTP_CONTROL_TYPE_IMAGE);
    globus_ftp_client_operationattr_set_mode(&ftp_attr,GLOBUS_FTP_CONTROL_MODE_EXTENDED_BLOCK);
    globus_ftp_client_operationattr_set_data_protection(&ftp_attr,GLOBUS_FTP_CONTROL_PROTECTION_PRIVATE);
    globus_ftp_client_operationattr_set_control_protection(&ftp_attr,GLOBUS_FTP_CONTROL_PROTECTION_PRIVATE);
    globus_gass_copy_attr_init(&o_attr);
    globus_gass_copy_attr_set_ftp(&o_attr,&ftp_attr);
  globus_mutex_init(&upload_lock,GLOBUS_NULL);
  globus_cond_init(&upload_cond,GLOBUS_NULL);
  upload_done=0; 
  if((error=globus_gass_copy_register_url_to_url(
               &handle,
               from_url,GLOBUS_NULL,
               to_url,&o_attr,
               &upload_finished_callback,NULL)) != GLOBUS_SUCCESS) {
      fprintf(stderr,"Failed to start upload\n");
      fprintf(stderr,"Error is: %s\n",globus_object_printable_to_string(globus_error_get(error)));
  }
  else while(!upload_done) { 
    globus_mutex_lock(&upload_lock);
    globus_cond_wait(&upload_cond,&upload_lock);
    globus_mutex_unlock(&upload_lock);
  };


  globus_module_deactivate(GLOBUS_FTP_CLIENT_DEBUG_PLUGIN_MODULE);
  globus_module_deactivate(GLOBUS_GASS_COPY_MODULE);
  return 0;
}
