/* Utility to duplicte file or directory from remote gridftp server */
#include "../std.h"
#include <stdlib.h>
#include <string>
#include <list>
extern "C" {
#include <globus_common.h>
#include <globus_error.h>
#include <globus_ftp_client.h>
#include <globus_gass_copy.h>
#include <globus_object.h>
#include "../misc/mkdir_recursive.h"
}
#include "../misc/canonical_dir.h"
#include "../files/info_types.h"
#include "../files/info_files.h"
#include "../misc/log_time.h"
#include "check.h"
#include "lister.h"
#include "ui_downloader.h"

#define GLOBUS_ERROR GLOBUS_ERROR_NO_INFO

globus_mutex_t c_lock;
globus_cond_t c_cond;
int c_done;

int url_to_path(string &s) {
  int n = s.find(':');
  if( n == string::npos) return -1;
  n++; if(n >= s.length()) return -1; if(s[n] != '/') return -1;
  n++; if(n >= s.length()) return -1; if(s[n] != '/') return -1;
  n++; n=s.find('/',n); if(n == string::npos) n=s.length();
  s.erase(0,n);
  return 0;
}

int list_recursively(const string &url,list<string> &names, const string &add_path = "") {
  static Lister *lister = NULL;
  static int depth;
  if(add_path.length() == 0) {
    depth=0;
    lister = new Lister;
    if((!lister) || (!(*lister))) {
      odlog(0)<<"Problem initializing Lister"<<endl;
      return 1;
    };
  }
  else {
    depth++;
    if(depth >= 20) return -1;
  };
  if(lister->retrieve_dir(url+add_path) != 0) {
    odlog(0)<<"List not retrieved: "<<url+add_path<<endl; return -1;
  };
  /* check if directory or file */
  if(lister->size() == 1) {
    string path=url+add_path;
    url_to_path(path);
    if(path == (*(lister->begin()))) { /* file */
      names.push_back(add_path);
      depth--;
      if(add_path.length() == 0) { delete lister; };
      return 0;
    };
  };
  /* directory(ies) */
  {
    list<string> tnames;
    for(list<string>::iterator i=lister->begin();i!=lister->end();++i) {
      tnames.push_back(*i);
    };
    for(list<string>::iterator i=tnames.begin();i!=tnames.end();++i) {
      string add_path_=add_path+'/'+(*i);
      list_recursively(url,names,add_path_);
    };
  };
  depth--;
  if(add_path.length() == 0) { delete lister; };
  return 0;
}

void deleted_callback(void* arg,globus_ftp_client_handle_t *handle,globus_object_t *error) {
  globus_mutex_lock(&c_lock);
  if(error != GLOBUS_SUCCESS) {
    odlog(0) << "Delete failed"<<endl;
    char* tmp = globus_object_printable_to_string(error);
    odlog(0)<<tmp<<endl; free(tmp);
    c_done=2;
  }
  else {
    odlog(2) << "File deleted"<<endl;
    c_done=1;
  };
  globus_cond_signal(&c_cond);
  globus_mutex_unlock(&c_lock);
}

void finished_callback(void* arg,globus_gass_copy_handle_t *handle,globus_object_t *error) {
  globus_mutex_lock(&c_lock);
  if(error != GLOBUS_SUCCESS) {
    odlog(0)<<"Transfer failed"<<endl;
    char* tmp = globus_object_printable_to_string(error);
    odlog(0)<<tmp<<endl; free(tmp);
    c_done=2;
  }
  else {
    odlog(2) << "File transfered"<<endl;
    c_done=1;
  };
  globus_cond_signal(&c_cond);
  globus_mutex_unlock(&c_lock);
}

#ifdef __UI_LIBRARY__
int ui_downloader(const char* url,bool recursive,const char* path,const vector<string> &filenames,bool download_files,bool remove_files,int debug_level) {
  char* session_url = url?strdup(url):NULL;
  LogTime::level=debug_level;
#else
int main(int argc,char* argv[]) {
  char* session_url = NULL;
  bool remove_files = true;
  bool download_files = true;
#endif
  bool gass_enabled = false;
  LogTime::active=false;
  int res = 1;
  if(globus_module_activate(GLOBUS_COMMON_MODULE) != GLOBUS_SUCCESS) {
    odlog(0)<<"COMMON module activation failed\n";
    if(session_url) free(session_url); return 1;
  };

  int l;
  char* path_ = NULL;
  bool download_successful = true;
#ifndef __UI_LIBRARY__
  const char* path = NULL;
  bool recursive = false;
  for(;;) {
    opterr=0;
    char optc=getopt(argc,argv,"+hkrp:");
    if(optc == -1) break;
    switch(optc) {
      case 'h': {
        cerr<<"ui-downloader [-r] [-k] [-p local_path] session_dir_url [list]"<<endl;
        exit(1);
      }; break;
      case 'p': {
        if(path != NULL) {
          cerr<<"Only one path can be given"<<endl;
          exit(1);
        };
        path=optarg;
      }; break;
      case 'r': {
        recursive=true;
      }; break;
      case 'k': {
        remove_files=false;
      }; break;
      case '?': {
        cerr<<"Unsupported option '"<<(char)optopt<<"'"<<endl;
        exit(1);
      }; break;
      case ':': {
        cerr<<"Missing parameter for option '"<<(char)optopt<<"'"<<endl;
        exit(1);
      }; break;
      default: {
        cerr<<"Undefined processing error"<<endl;
        exit(1);
      };
    };
  };
  session_url = strdup(argv[optind]);
#endif
  if(!session_url) {
    odlog(0) << "Missing session directory url" << endl; return 1;
  };
  l = strlen(session_url);
  if(l == 0) {
    odlog(0) << "Wrong session url" << endl; free(session_url); return 1;
  };
  if(session_url[l-1] == '/') session_url[l-1]=0;
#ifndef __UI_LIBRARY__
  if(argv[optind+1] && recursive) {
    cerr << "Recursive and listed retrieval can't be requested at same time" << endl;
    free(session_url); return 1;
  };
#endif
  if(!path) path="";
  l = strlen(path);
  if(path[0] != '/') {
    path_=(char*)malloc(1024+l+1); 
    if(path_ == NULL) { odlog(0)<<"Memory allocation error"<<endl; return 1; };
    memset(path_,0,1024); getcwd(path_,1023); 
    strcat(path_,"/"); strcat(path_,path); path=path_;
  };
  l = strlen(path);
  if(path[l-1] == '/') ((char*)path)[l-1]=0;
  odlog(1)<<"Remote base url: "<<session_url<<endl;
  odlog(1)<<"Local path: "<<path<<endl;

  list<FileData> job_files;
  if(recursive) {
    list<string> names;
    if(globus_module_activate(GLOBUS_FTP_CONTROL_MODULE) != GLOBUS_SUCCESS) {
      odlog(0)<<"FTP CONTROL module activation failed\n";
      globus_module_deactivate(GLOBUS_COMMON_MODULE);
      res=1; goto exit;
    };
    odlog(1)<<"Listing available files: "<<endl;
    if(list_recursively(string(session_url),names) != 0) {
      odlog(0)<<"List failed"<<endl; res=1;
      globus_module_deactivate(GLOBUS_FTP_CONTROL_MODULE);
      goto exit;
    }
    else {
      for(list<string>::iterator i=names.begin();i!=names.end();++i) {
        job_files.push_back(FileData(i->c_str(),NULL));
        odlog(1)<<"  "<<(*i)<<endl;
      };
    };
    globus_module_deactivate(GLOBUS_FTP_CONTROL_MODULE);
  }
  else {
#ifndef __UI_LIBRARY__
    char* file_list=argv[optind+1];
    if(file_list) {
      if(!job_Xput_read_file(string(file_list),job_files)) {
        cerr << "Can't read list of files from " << file_list << endl;
        res=1; goto exit;
      };
    }
    else {
      if(!job_Xput_read_file(job_files)) {
        cerr << "Can't parse list of files" << endl;
        res=1; goto exit;
      };
    };
#else
  for(vector<string>::const_iterator iv=filenames.begin();iv!=filenames.end();++iv) {
    string f1 = *iv;
    string f2;
    canonical_dir(f1);
    ++iv; if(iv==filenames.end()) {
      job_files.push_back(FileData(f1.c_str(),NULL));
      break;
    };
    f2 = *iv;
    job_files.push_back(FileData(f1.c_str(),f2.c_str()));
  };
#endif
  };
  if(job_files.size() == 0) {
    odlog(1)<<"No files to download"<<endl; res=0; goto exit;
  };

  if(globus_module_activate(GLOBUS_GASS_COPY_MODULE) != GLOBUS_SUCCESS) {
    odlog(0)<<"GASS COPY module activation failed\n";
    goto exit;
  };
  gass_enabled = true;
  globus_mutex_init(&c_lock,GLOBUS_NULL);
  globus_cond_init(&c_cond,GLOBUS_NULL);
  if(download_files) {
  /* download */
  globus_gass_copy_handle_t c_h;
  globus_gass_copy_handle_init(&c_h,GLOBUS_NULL);
  globus_gass_copy_attr_t c_attr;
  globus_ftp_client_operationattr_t c_ftp_attr;
  globus_gass_copy_attr_init(&c_attr);
  globus_ftp_client_operationattr_init(&c_ftp_attr);
  globus_ftp_client_operationattr_set_data_protection(&c_ftp_attr,
                                 GLOBUS_FTP_CONTROL_PROTECTION_PRIVATE);
  globus_gass_copy_attr_set_ftp(&c_attr,&c_ftp_attr);
  string rurl;
  string lurl;
  string lfile;
  string ldir;
  c_done=0;
  for(FileData::iterator i=job_files.begin();i!=job_files.end();) {
    if(i->lfn.length() != 0) { lfile=i->lfn; } else { lfile=i->pfn; };
    if( lfile[0] != '/' ) { lfile=string(path)+"/"+lfile; }
    else { lfile=string(path)+lfile; };
    lurl="file://"+lfile;
    if(i->pfn[0] != '/') { rurl=string(session_url)+'/'+i->pfn; }
    else { rurl=string(session_url)+i->pfn; };
    ldir=lfile; int n = ldir.rfind('/'); if(n == string::npos) n=0;
    ldir.erase(n,string::npos);

    if(mkdir_recursive("",ldir.c_str(),0700,getuid(),getgid()) != 0) {
      if(errno != EEXIST) {
        odlog(1)<<"Failed to create directory "<<ldir<<endl;
        globus_gass_copy_handle_destroy(&c_h);
        globus_ftp_client_operationattr_destroy(&c_ftp_attr);
        free(session_url); goto exit;
      };
    };

    odlog(1)<<"Downloading "<<rurl<<"\n  to "<<lfile<<endl;

    bool downloaded = false;

    if((globus_object_t*)globus_gass_copy_register_url_to_url_check(
               &c_h,
               (char*)(rurl.c_str()),&c_attr,
               (char*)(lurl.c_str()),GLOBUS_NULL,
               &finished_callback,NULL) != GLOBUS_SUCCESS) {
      odlog(0) << "Probably bad url or local file: " <<rurl<<"  "<<lurl<< endl;
    }
    else {
      globus_mutex_lock(&c_lock);
      while(!c_done) {
        globus_cond_wait(&c_cond,&c_lock);
      };
      downloaded=(c_done == 1); 
      c_done=0;
      globus_mutex_unlock(&c_lock);
    };
    fix_file_permissions(lfile,false);
    if(downloaded) { i++; }
    else {
      i=job_files.erase(i);
      download_successful = false;
    };
  };
  globus_gass_copy_handle_destroy(&c_h);
  globus_ftp_client_operationattr_destroy(&c_ftp_attr);
  };
  if(remove_files) {
  /* delete */
  string rurl;
  globus_ftp_client_handle_t c_h;
  globus_ftp_client_handle_init(&c_h,GLOBUS_NULL);
  c_done=0;
  for(FileData::iterator i=job_files.begin();i!=job_files.end();++i) {
    if(i->pfn[0] != '/') { rurl=string(session_url)+'/'+i->pfn; }
    else { rurl=string(session_url)+i->pfn; };
    odlog(1)<<"Deleting "<<rurl<<endl;
    if(globus_ftp_client_delete(&c_h,rurl.c_str(),NULL,&deleted_callback,NULL)
              != GLOBUS_SUCCESS) {
      odlog(0) << "Probaly bad url: "<<rurl<< endl;
    }
    else {
      globus_mutex_lock(&c_lock);
      while(!c_done) {
        globus_cond_wait(&c_cond,&c_lock);
      };
      c_done=0;
      globus_mutex_unlock(&c_lock);
    };
  };
  globus_ftp_client_handle_destroy(&c_h);
  };
  globus_cond_destroy(&c_cond);
  globus_mutex_destroy(&c_lock);
  if(download_successful) res = 0;
exit:
  free(session_url); if(path_) free(path_);
  if(gass_enabled) globus_module_deactivate(GLOBUS_GASS_COPY_MODULE);
  globus_module_deactivate(GLOBUS_COMMON_MODULE);
  return res;
}
