#include "std.h"

#include <iostream>
#include <string>
#include <fstream>
#include <vector>
#include <stdio.h>

#include "Preferences.h"

string GetEnv (const string & env) {

  const char * temp = getenv (env.c_str());
  return temp ? temp : "";
}

string GetDef (const string & def) {

  static bool doinit = true;
  static vector <string> defaults;

  if (doinit) {
    string conffilename = GetEnv ("HOME");
    conffilename.append ("/.ngrc");
    ifstream conffile (conffilename.c_str());
    string line;
    while (getline (conffile, line)) {
      if (line.empty()) continue;
      if (line[0] == '#') continue;
      int pos = line.find ('=');
      if (pos == string::npos)
	cerr << "Syntax error in " << conffilename << ": " << line << endl;
      else {
	int quot = 0;
	if (line[pos + 1] == '"' &&
	    line[line.length() - 1] == '"') quot = 1;
	if (line[pos + 1] == '\'' &&
	    line[line.length() - 1] == '\'') quot = 1;
	defaults.push_back (line.substr (0, pos));
	defaults.push_back (line.substr (pos + 1 + quot,
					 line.length() - pos - 1 - 2 * quot));
      }
    }
    conffile.close();
    doinit = false;
  }
  for (vector <string>::iterator it = defaults.begin();
       it != defaults.end(); it++, it++)
    if (*it == def)
      return *(++it);
  return "";
}

int iGetEnv (const string & env) {

  string temp = GetEnv (env);
  if (temp.empty())
    return UNDEFINED;
  else
    return atoi (temp.c_str());
}

int iGetDef (const string & def) {

  string temp = GetDef (def);
  if (temp.empty())
    return UNDEFINED;
  else
    return atoi (temp.c_str());
}

vector <string> GetJobIDs (int argc, char ** argv, int optind,
			   vector <string> & joblist, bool all) {

  vector <string> jobids;

  string filename = GetEnv ("HOME");
  filename.append ("/.ngjobs");
  ifstream ngjobs (filename.c_str());

  if (all) {
    string jobidname;
    while (getline (ngjobs, jobidname)) {
      int pos = jobidname.find('#');
      string jobid = jobidname.substr (0, pos);
      string jobname = jobidname.substr (pos + 1);
      bool old = false;
      for (vector <string>::iterator vsi = jobids.begin();
	   !old && vsi != jobids.end(); vsi++)
	if (*vsi == jobid) old = true;
      if (!old) jobids.push_back (jobid);
    }
  }

  for (int i = optind; i < argc; i++) {
    bool found = false;
    ngjobs.seekg (ios::beg);
    ngjobs.clear();
    string jobidname;
    while (getline (ngjobs, jobidname)) {
      int pos = jobidname.find('#');
      string jobid = jobidname.substr (0, pos);
      string jobname = jobidname.substr (pos + 1);
      if (jobid == argv[i] || jobname == argv[i]) {
	bool old = false;
	for (vector <string>::iterator vsi = jobids.begin();
	     !old && vsi != jobids.end(); vsi++)
	  if (*vsi == jobid) old = true;
	if (!old) jobids.push_back (jobid);
	found = true;
      }
    }
    if (!found) {
      bool old = false;
      for (vector <string>::iterator vsi = jobids.begin();
	   !old && vsi != jobids.end(); vsi++)
	if (*vsi == argv[i]) old = true;
      if (!old) jobids.push_back (argv[i]);
    }
  }

  ngjobs.close();

  if (!joblist.empty()) {
    for (vector <string>::iterator vsi = joblist.begin();
	   vsi != joblist.end(); vsi++) {
      ifstream joblist (vsi->c_str());
      if (!joblist) cerr << "Warning: can't open file " << *vsi << endl;
      string jobid;
      while (getline (joblist, jobid)) {
	bool old = false;
	for (vector <string>::iterator vsi = jobids.begin();
	     !old && vsi != jobids.end(); vsi++)
	  if (*vsi == jobid) old = true;
	if (!old) jobids.push_back (jobid);
      }
      joblist.close();
    }
  }

  return jobids;
}

void RemoveJobID (const string & jobid) {

  string oldfilename = GetEnv ("HOME");
  oldfilename.append ("/.ngjobs");
  ifstream oldngjobs (oldfilename.c_str());
  string newfilename = oldfilename + ".tmp";
  ofstream newngjobs (newfilename.c_str());
  string jobidname;
  while (getline (oldngjobs, jobidname)) {
    int pos = jobidname.find('#');
    string jobid_ = jobidname.substr (0, pos);
    if (jobid_ != jobid) newngjobs << jobidname << endl;
  }
  oldngjobs.close();
  newngjobs.close();
  remove (oldfilename.c_str());
  rename (newfilename.c_str(), oldfilename.c_str());
}
